/*
 * Interpolator.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include "stdafx.h"

#include "Interpolator.h"
#include "FieldValue.h"
#include "MFFloat.h"
#include "Scene.h"
#include "MFieldCommand.h"
#include "FieldCommand.h"
#include "CommandList.h"

ProtoInterpolator::ProtoInterpolator(Scene *scene, const char *name, int keyType,
					 int keysType, FieldValue *defaultValue)
  : Proto(scene, name)
{
    addEventIn(SFFLOAT, "set_fraction");
    addExposedField(MFFLOAT, "key", new MFFloat());
    addExposedField(keysType, "keyValue", defaultValue);
    addEventOut(keyType, "value_changed");
}

Interpolator::Interpolator(Scene *scene, Proto *def)
  : Node(scene, def)
{
    _fraction = 0.0f;
}

int
Interpolator::getNumKeys() const
{
    MFFloat	*key = (MFFloat *) getField(0);

    return key->getSize();
}

float
Interpolator::getKey(int index) const
{
    MFFloat	*key = (MFFloat *) getField(0);
    return key->getValue(index);
}

float
Interpolator::getKeyValue(int channel, int index) const
{
    MFFloat	*keyValue = (MFFloat *) getField(1);
    return keyValue->getValue(index * getNumChannels() + channel);
}

void
Interpolator::setKey(int index, float value)
{
    MFFloat	*key = (MFFloat *) getField(0);

    key->setValue(index, value);
    _scene->OnFieldChange(this, 0, index);
}

void
Interpolator::setKeyValue(int channel, int index, float value)
{
    MFFloat	*keyValue = (MFFloat *) getField(1);

    keyValue->setValue(index * getNumChannels() + channel, value);
    _scene->OnFieldChange(this, 1, index);
}


void
Interpolator::backupKey(int index)
{
    CommandList	    *list = new CommandList();

    list->append(new MFieldCommand(this, 0, index));
    list->append(new MFieldCommand(this, 1, index));
    _scene->add(list);
}

void
Interpolator::backup(CommandList *list)
{
    int	    pos = findKeyInclusive(_fraction);

    if (pos != getNumKeys()) {
	list->append(new MFieldCommand(this, 0, pos));
	list->append(new MFieldCommand(this, 1, pos));
    }
    list->append(new FieldCommand(this, 0));
    list->append(new FieldCommand(this, 1));
}

//
// findKey() - return the closest key stricty greater than the given value,
//	       or n if none found
//

int
Interpolator::findKey(float value) const
{
    MFFloat    *key = (MFFloat *) getField(0);
    int		i, numKeys = key->getSize();

    for(i = 0; i < numKeys; i++) {
	float	    k = key->getValue(i);
	if (k > value) break;
    }
    return i;
}

//
// findKeyInclusive() - return the closest key greater than or equal to 
//			the given value, or n if none found
//

int
Interpolator::findKeyInclusive(float value) const
{
    MFFloat    *key = (MFFloat *) getField(0);
    int		i, numKeys = key->getSize();

    for(i = 0; i < numKeys; i++) {
	float	    k = key->getValue(i);
	if (k >= value) break;
    }
    return i;
}

bool
Interpolator::getNearestKeys(float k, float *k1, float *k2, int *pos1, int *pos2)
{
    MFFloat    *key = (MFFloat *) getField(0);
    int		numKeys = key->getSize();

    if (numKeys == 0) {
	return false;
    }

    int pos = findKey(k);

    if (pos == 0) {
        *k1 = 0.0f;
	*pos1 = 0;
    } else {
	*k1 = key->getValue(pos - 1);
	*pos1 = pos - 1;
    }

    if (pos == numKeys) {
	*k2 = 1.0f;
	*pos2 = numKeys-1;
    } else {
	*k2 = key->getValue(pos);
	*pos2 = pos;
    }
    return true;
}

void
Interpolator::receiveEvent(int eventIn, double timestamp, FieldValue *value)
{
    switch (eventIn) {
      case 0:
	sendInterpolatedValue(timestamp, ((SFFloat *) value)->getValue());
        break;
      default:
	Node::receiveEvent(eventIn, timestamp, value);
	break;
    }
}

void
Interpolator::sendInterpolatedValue(double timestamp, float k)
{
    _fraction = k;
    float      *values = new float[getNumChannels()];
    interpolate(k, values);
    sendEvent(2, timestamp, createKey(values));
    delete [] values;
}

void
Interpolator::interpolate(float k, float *values)
{
    MFFloat    *keyValue = (MFFloat *) getField(1);
    int		numChannels = getNumChannels();

    float   k1, k2;
    int	    pos1, pos2;

    if (keyValue->getSize() == 0) {
	for (int i = 0; i < numChannels; i++) {
	    values[i] = 0.0f;
	}
	return;
    }

    getNearestKeys(k, &k1, &k2, &pos1, &pos2);

    const float    *value = keyValue->getValue();

    float	    alpha;

    if (k1 == k2) {
	alpha = 0.0;
    } else {
	alpha = (k - k1) / (k2 - k1);
    }

    for (int i = 0; i < numChannels; i++) {
	float	    v1 = value[pos1 * numChannels + i];
	float	    v2 = value[pos2 * numChannels + i];
	values[i] = v1 + (v2 - v1) * alpha;
    }
}

void
Interpolator::insertKey(int pos, float value, const float *values)
{
    MFFloat        *key = (MFFloat *) getField(0);
    MFFloat	   *keyValue = (MFFloat *) getField(1);
    int		    numKeys = key->getSize();
    int		    numChannels = getNumChannels();
    const float    *k = key->getValue();
    const float	   *v = keyValue->getValue();

    float	   *newK = new float[numKeys + 1];
    float	   *newV = new float[(numKeys + 1) * numChannels];

    if (pos > 0) {
	memcpy(newK, k, pos * sizeof(float));
	memcpy(newV, v, pos * numChannels * sizeof(float));
    }

    if (pos < numKeys) {
	memcpy(newK + pos + 1, k + pos, (numKeys - pos) * sizeof(float));
	memcpy(newV + (pos + 1) * numChannels, v + pos * numChannels, 
	       (numKeys - pos) * numChannels * sizeof(float));
    }

    newK[pos] = value;
    if (values) {
	for (int i = 0; i < numChannels; i++) {
	    newV[pos * numChannels + i] = values[i];
	}
    }

    MFFloat	    *newKey = new MFFloat(newK, numKeys + 1);
    FieldValue	    *newKeyValue = createKeys(newV, numKeys + 1);
    
    delete [] newK;
    delete [] newV;

    if (values) {
	CommandList	    *list = new CommandList();

	list->append(new FieldCommand(this, 0, newKey));
	list->append(new FieldCommand(this, 1, newKeyValue));
	_scene->execute(list);
    } else {
	setField(0, newKey);
	setField(1, newKeyValue);
	_scene->OnFieldChange(this, 0);
	_scene->OnFieldChange(this, 1);
    }
}

void
Interpolator::deleteKeys(int start, int end)
{
    MFFloat        *key = (MFFloat *) getField(0);
    MFFloat	   *keyValue = (MFFloat *) getField(1);
    int		    numKeys = key->getSize();
    int		    numChannels = getNumChannels();
    const float    *k = key->getValue();
    const float	   *v = keyValue->getValue();

    if (start == end) return;

    int		    newlen = numKeys - (end - start);
    float	   *newK = new float[newlen];
    float	   *newV = new float[newlen * numChannels];

    if (start > 0) {
	memcpy(newK, k, start * sizeof(float));
	memcpy(newV, v, start * numChannels * sizeof(float));
    }

    if (end < numKeys) {
	memcpy(newK + start, k + end, (numKeys - end) * sizeof(float));
	memcpy(newV + start * numChannels, v + end * numChannels, 
	       (numKeys - end) * numChannels * sizeof(float));
    }

    MFFloat	    *newKey = new MFFloat(newK, newlen);
    FieldValue	    *newKeyValue = createKeys(newV, newlen);

    delete [] newK;
    delete [] newV;

    CommandList	    *list = new CommandList();

    list->append(new FieldCommand(this, 0, newKey));
    list->append(new FieldCommand(this, 1, newKeyValue));
    _scene->execute(list);
}

//
// recordKey() -- record a key at the current position (_fraction)
//

void
Interpolator::recordKey(FieldValue *value)
{
    int		 key = findKeyInclusive(_fraction);

    if (key == getNumKeys() || !EQUALF(getKey(key), _fraction)) {
	// no exact key found, create one

	insertKey(key, _fraction, NULL);
    }
    MFFloat	*keyValue = (MFFloat *) getField(1);
    keyValue->setSFValue(key, value);
    _scene->OnFieldChange(this, 0);
    _scene->OnFieldChange(this, 1);
}

