/*
 * Node.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "stdafx.h"

#include "Node.h"
#include "Scene.h"
#include "Field.h"
#include "FieldValue.h"
#include "FieldCommand.h"
#include "Proto.h"
#include "Path.h"
#include "SFNode.h"
#include "MFNode.h"
#include "SFString.h"
#include "MFString.h"
#include "EventIn.h"
#include "EventOut.h"
#include "ExposedField.h"
#include "URL.h"

#define HANDLE_SIZE 2.0f

Node::Node(Scene *scene, Proto *proto)
{
    _scene = scene;
    _graphX = _graphY = 0;
    _graphWidth = _graphHeight = 0;
    _scene->addNode(this);
    _refs = 0;
    _flags = 1<<NODE_FLAG_COLLAPSED;
    _proto = proto;

    int	    i;
    _numFields = _proto->getNumFields();
    _fields = new FieldValue *[_numFields];
    for (i = 0; i < _numFields; i++) {
	_fields[i] = _proto->getField(i)->getDefault();
	_fields[i]->ref();
    }
    _numEventIns = _proto->getNumEventIns();
    _inputs = new SocketList [_numEventIns];
    _numEventOuts = _proto->getNumEventOuts();
    _outputs = new SocketList [_numEventOuts];
}

Node::Node(const Node &node)
{
    _scene = node._scene;
    _refs = 0;
    _graphX = _graphY = 0;
    _graphWidth = _graphHeight = 0;
    _scene->addNode(this);
    _flags = node._flags;
    _proto = node._proto;
    _numFields = _proto->getNumFields();
    _fields = new FieldValue *[_numFields];
    _scene->def(_name, this);

    for (int i = 0; i < _numFields; i++) {
	_fields[i] = node._fields[i]->copy();
	_fields[i]->ref();
    }
    _numEventIns = _proto->getNumEventIns();
    _inputs = new SocketList [_numEventIns];
    _numEventOuts = _proto->getNumEventOuts();
    _outputs = new SocketList [_numEventOuts];
}

Node::~Node()
{
    _scene->removeNode(this);
    for (int i = 0; i < _numFields; i++) {
	_fields[i]->unref();
    }
    delete [] _fields;
    delete [] _inputs;
    delete [] _outputs;

    _scene->undef(_name);
}

FieldValue *
Node::getField(int index) const
{
    assert (index >= 0 && index < _numFields);

    return _fields[index];
}

void
Node::setField(int index, FieldValue *value)
{
    assert (index >= 0 && index < _numFields);
    assert( _fields[index]->getType() == value->getType() );

    // if field is an SFNode or MFNode type, remove old values from 
    // children's parent list

    if (value->getType() == SFNODE) {
	Node *child = ((SFNode *) _fields[index])->getValue();
	if (child) child->removeParent(this, index);
    } else if (value->getType() == MFNODE) {
	NodeList *childList = ((MFNode *) _fields[index])->getValue();
	if (childList) {
	    for (int i = 0; i < childList->size(); i++) {
		Node *child = childList->get(i);
		if (child) child->removeParent(this, index);
	    }
	}
    }

    value->clamp(_proto->getField(index)->getMin(), 
		 _proto->getField(index)->getMax());

    _fields[index]->unref();
    _fields[index] = value;
    _fields[index]->ref();

    if (value->getType() == SFNODE) {
	Node *child = ((SFNode *) value)->getValue();
	if (child) child->addParent(this, index);
    } else if (value->getType() == MFNODE) {
	NodeList *childList = ((MFNode *) value)->getValue();
	if (childList) {
	    for (int i = 0; i < childList->size(); i++) {
		Node *child = childList->get(i);
		if (child) child->addParent(this, index);
	    }
	}
    }
}

void
Node::write(FILE *f, int indent)
{
    if (_proto) {
	indentf(f, indent);
	if (getFlag(NODE_FLAG_TOUCHED)) {
	    fprintf(f, "USE %s\n", (const char *) _name);
	} else {
	    if (needsDEF()) {
		if (!_name[0]) _scene->generateUniqueNodeName(this);
		fprintf(f, "DEF %s ", (const char *) _name);
	    }
	    fprintf(f, "%s {\n", (const char *) _proto->getName());
	    writeFields(f, indent+2);
	    writeRoutes(f, indent+2);
	    indentf(f, indent);
	    fprintf(f, "}\n");
	    setFlag(NODE_FLAG_TOUCHED);
	}
    }
}

bool
Node::needsDEF() const
{
    if (getNumParents() > 1) return true;

    for (int i = 0; i < _numEventIns; i++) {
	if (_inputs[i].size() > 0) {
	    return true;
	}
    }

    for (int j = 0; j < _numEventOuts; j++) {
	if (_outputs[j].size() > 0) {
	    return true;
	}
    }

    return false;
}

void
Node::writeFields(FILE *f, int indent)
{
    if (!_proto) return;
    const char	   *oldBase = _scene->getURL();
    const char	   *newBase = _scene->getNewURL();
    bool	    tempSave = _scene->isTempSave();

    for (int i = 0; i < _numFields; i++) {
	Field		*field = _proto->getField(i);
	FieldValue	*value = _fields[i];
	if (value && !value->equals(field->getDefault())) {
	    indentf(f, indent);
	    fprintf(f, "%s ", (const char *) field->getName());
	    if (field->getFlags() & FF_URL) {
		value = rewriteField(value, oldBase, newBase);
		value->write(f, indent);
		if (!tempSave) {
		    setField(i, value);
		} else {
		    delete value;
		}
	    } else {
		value->write(f, indent);
	    }
	}
    }
}

FieldValue *
Node::rewriteField(FieldValue *value, const char *oldBase, const char *newBase)
{
    String	r;

    switch (value->getType()) {
      case SFSTRING:
	r = rewriteURL(((SFString *) value)->getValue(), oldBase, newBase);
	return new SFString(r);
      case MFSTRING:
      {
	StringArray *a = new StringArray();
	int n = ((MFString *) value)->getSize();
	for (int i = 0; i < n; i++) {
	    (*a)[i] = rewriteURL(((MFString *) value)->getValue(i), oldBase, newBase);
	}
	return new MFString(a);
      }
      default:
	assert( 0 );
	return NULL;
    }
}

String
Node::rewriteURL(const char *url, const char *oldBase, const char *newBase)
{
    URL		oldURL(oldBase, url);
    String	newURL(oldURL.RelativeTo(newBase));

    return newURL;
}

void
Node::writeRoutes(FILE *f, int indent) const
{
    int				i;
    SocketList::Iterator       *j;

    if (!_proto) return;

    for (i = 0; i < _numEventIns; i++) {
	for (j = _inputs[i].first(); j != NULL; j = j->next()) {
	    if (j->item()._node->getFlag(NODE_FLAG_TOUCHED)) {
		indentf(f, indent);
		Node	*src = j->item()._node;
		fprintf(f, "ROUTE %s.%s TO %s.%s\n",
			(const char *) src->getName(),
			(const char *) src->getProto()->getEventOut(j->item()._index)->getName(),
			(const char *) _name,
			(const char *) getProto()->getEventIn(i)->getName());
	    }
	}
    }

    for (i = 0; i < _numEventOuts; i++) {
	for (j = _outputs[i].first(); j != NULL; j = j->next()) {
	    if (j->item()._node->getFlag(NODE_FLAG_TOUCHED)) {
		indentf(f, indent);
		Node	*dst = j->item()._node;
		fprintf(f, "ROUTE %s.%s TO %s.%s\n",
			(const char *) _name,
			(const char *) getProto()->getEventOut(i)->getName(),
			(const char *) dst->getName(),
			(const char *) dst->getProto()->getEventIn(j->item()._index)->getName());
	    }
	}
    }
}

int
Node::lookupEventIn(const String &name) const
{
    return _proto ? _proto->lookupEventIn(name) : INVALID_INDEX;
}

int
Node::lookupEventOut(const String &name) const
{
    return _proto ? _proto->lookupEventOut(name) : INVALID_INDEX;
}

void
Node::addInput(int eventIn, Node *src, int eventOut)
{
    _inputs[eventIn].append(Socket(src, eventOut));
}

void
Node::addOutput(int eventOut, Node *dst, int eventIn)
{
    _outputs[eventOut].append(Socket(dst, eventIn));
}

void
Node::removeInput(int eventIn, Node *src, int eventOut)
{
    SocketList::Iterator	*j = _inputs[eventIn].find(Socket(src, eventOut));
    if (j) 
	_inputs[eventIn].remove(j);
}

void
Node::removeOutput(int eventOut, Node *dst, int eventIn)
{
    SocketList::Iterator	*j = _outputs[eventOut].find(Socket(dst, eventIn));
    if (j)
	_outputs[eventOut].remove(j);
}

void
Node::update()
{
    // used by Script to update its fields
}

void
Node::addParent(Node *parent, int field)
{
    _parents.append(Socket(parent, field));
}

void
Node::removeParent(Node *parent, int field)
{
    SocketList::Iterator *i = _parents.find(Socket(parent, field));

    if (!i) {
	assert(false);
    } else {
	_parents.remove(i);
    }
}

int
Node::findChild(Node *child, int field) const
{
    assert(field >= 0 && field < _numFields);

    FieldValue	*value = _fields[field];

    if (value->getType() == SFNODE) {
	return (((SFNode *) value)->getValue() == child) ? 0 : -1;
    } else if (value->getType() == MFNODE) {
	NodeList    *list = ((MFNode *) value)->getValue();
	for (int i = 0; i < list->size(); i++) {
	    if (list->get(i) == child) return i;
	}
	return -1;
    } else {
	return -1;
    }
}

bool
Node::hasAncestor(Node *node) const
{    
    for (SocketList::Iterator *i = _parents.first(); i != NULL; i = i->next()) {
	Node	*parent = i->item()._node;
	if (parent == node || parent->hasAncestor(node)) return true;
    }
    return false;
}

int
Node::findValidFieldType(int childType)
{
    int match = -1;
    for (int field = 0; field < _numFields; field++) {
	if (validChildType(field, childType)) {
	    if (match != -1) return -1;	    // ambiguous
	    match = field;
	}
    }
    return match;
}

int
Node::findValidField(Node *child)
{
    return findValidFieldType(child->getNodeClass());
}

bool
Node::validChild(int field, Node *child)
{
    return validChildType(field, child->getNodeClass());
}

bool
Node::validChildType(int field, int childType)
{
    if (field < 0 || field >= _numFields) return false;

    Field	    *def = _proto->getField(field);

    if (def->getType() == SFNODE && ((SFNode *) _fields[field])->getValue() != NULL)
	return false;

    if (def->getType() == SFNODE || def->getType() == MFNODE) {
	int	nodeType = def->getNodeType();

	if (nodeType == NODE_ANY) {
	    return true;
	} else if (childType == NODE_MOVIE_TEXTURE && 
		   (nodeType == NODE_AUDIO_CLIP || nodeType == NODE_TEXTURE)) {
	    return true;
	} else {
	    return childType == nodeType;
	}
    } else {
	return false;
    }
}

void
Node::drawHandle(const Vec3f &pos)
{
    float	wx, wy, wz;
    float   f[4];
    _scene->projectPoint(pos.x, pos.y, pos.z, &wx, &wy, &wz);
    glGetFloatv(GL_VIEWPORT, f);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(f[0], f[0] + f[2], f[1], f[1] + f[3]);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    glBegin(GL_POLYGON);
    glVertex3f(wx - HANDLE_SIZE, wy - HANDLE_SIZE, wz);
    glVertex3f(wx - HANDLE_SIZE, wy + HANDLE_SIZE, wz);
    glVertex3f(wx + HANDLE_SIZE, wy + HANDLE_SIZE, wz);
    glVertex3f(wx + HANDLE_SIZE, wy - HANDLE_SIZE, wz);
    glEnd();

    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
}

void
Node::setFlagRec(int flag)
{
    setFlag(flag);

    for (int i = 0; i < _numFields; i++) {
	if (_fields[i]->getType() == MFNODE) {
	    ((MFNode *) _fields[i])->getValue()->setFlag(flag);
	} else if (_fields[i]->getType() == SFNODE) {
	    SFNode *value = (SFNode *) _fields[i];
	    if (value->getValue()) value->getValue()->setFlagRec(flag);
	}
    }
}

void
Node::clearFlagRec(int flag)
{
    setFlag(flag);

    for (int i = 0; i < _numFields; i++) {
	if (_fields[i]->getType() == MFNODE) {
	    ((MFNode *) _fields[i])->getValue()->clearFlag(flag);
	} else if (_fields[i]->getType() == SFNODE) {
	    SFNode *value = (SFNode *) _fields[i];
	    if (value->getValue()) value->getValue()->clearFlagRec(flag);
	}
    }
}

Path *
Node::getPrimaryPath() const
{
    int		len = 0;
    const Node       *node, *root = _scene->getRoot();

    for (node = this; node != root; node = node->getParent(0)) {
	len += 2;
    }

    int	   *list = new int[len];

    int i = len-1;
    
    for (node = this; node != root; node = node->getParent(0)) {
	Node	   *parent = node->getParent(0);
	int	    field = node->getParentField(0);
	FieldValue *value = parent->getField(field);
	if (value->getType() == MFNODE) {
	    list[i--] = ((MFNode *) value)->getValue()->find((Node *) node);
	} else if (value->getType() == SFNODE)  {
	    list[i--] = 0;
	} else {
	    assert(false);
	}
	list[i--] = field;
    }

    Path    *path = new Path(list, len, _scene);
    delete [] list;
    return path;
}

bool
Node::isInScene() const
{
    if (this == _scene->getRoot()) return true;

    SocketList::Iterator    *i;
    for (i = _parents.first(); i != NULL; i = i->next()) {
	if (i->item()._node->isInScene())
	    return true;
    }
    return false;
}

void
Node::sendEvent(int eventOut, double timestamp, FieldValue *value)
{
    assert(eventOut >= 0 && eventOut <= _numEventOuts);

    value->ref();

    SocketList::Iterator    *i;

    for (i = _outputs[eventOut].first(); i != NULL; i = i->next()) {
	Socket	s = i->item();

	s._node->receiveEvent(s._index, timestamp, value);
    }

    value->unref();
}

void
Node::receiveEvent(int eventIn, double /* timestamp */, FieldValue *value)
{
    // check to see if this eventIn is part of an exposedField

    ExposedField       *e = _proto->getEventIn(eventIn)->getExposedField();

    if (e) {
	// set the appropriate field
	setField(e->getField(), value);
	_scene->OnFieldChange(this, e->getField());

	// fire off an event here?
    }
}
