/*
 * NodeElevationGrid.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <GL/gl.h>
#include "stdafx.h"

#include "NodeElevationGrid.h"
#include "Proto.h"
#include "FieldValue.h"
#include "SFNode.h"
#include "SFFloat.h"
#include "MFFloat.h"
#include "SFBool.h"
#include "SFInt32.h"
#include "MFVec3f.h"
#include "Scene.h"
#include "FieldCommand.h"

#define HEIGHT(i, j)	(height[(i) + (j) * xDimension])

ProtoElevationGrid::ProtoElevationGrid(Scene *scene)
  : Proto(scene, "ElevationGrid")
{
    addEventIn(MFFLOAT, "set_height");
    addExposedField(SFNODE, "color", new SFNode(NULL), NODE_COLOR);
    addExposedField(SFNODE, "normal", new SFNode(NULL), NODE_NORMAL);
    addExposedField(SFNODE, "texCoord", new SFNode(NULL), NODE_TEXTURE_COORDINATE);
    addField(MFFLOAT, "height", new MFFloat());
    addField(SFBOOL, "ccw", new SFBool(true));
    addField(SFBOOL, "colorPerVertex", new SFBool(true));
    addField(SFFLOAT, "creaseAngle", new SFFloat(0.0f), 0.0f);
    addField(SFBOOL, "normalPerVertex", new SFBool(true));
    addField(SFBOOL, "solid", new SFBool(true));
    addField(SFINT32, "xDimension", new SFInt32(0), 0.0f);
    addField(SFFLOAT, "xSpacing", new SFFloat(1.0f), 0.0f);
    addField(SFINT32, "zDimension", new SFInt32(0), 0.0f);
    addField(SFFLOAT, "zSpacing", new SFFloat(1.0f), 0.0f);
}

NodeElevationGrid::NodeElevationGrid(Scene *scene, Proto *def)
  : Node(scene, def)
{
    _normals = NULL;
    _oldX = _oldZ = 0;
}

NodeElevationGrid::~NodeElevationGrid()
{
    if (_normals != NULL) _normals->unref();
}

void
NodeElevationGrid::draw()
{
//    Node	 *color = ((SFNode *) getField(0))->getValue();
    Node	 *normal = ((SFNode *) getField(1))->getValue();
//    Node	 *texCoord = ((SFNode *) getField(2))->getValue();
    const float  *height = ((MFFloat *) getField(3))->getValue();
    int		  ccw = ((SFBool *) getField(4))->getValue();
//    int		  colorPerVertex = ((SFBool *) getField(5))->getValue();
//    float	  creaseAngle = ((SFFloat *) getField(6))->getValue();
    bool	  normalPerVertex = ((SFBool *) getField(7))->getValue();
    int		  solid = ((SFBool *) getField(8))->getValue();
    int		  xDimension = ((SFInt32 *) getField(9))->getValue();
    float	  xSpacing = ((SFFloat *) getField(10))->getValue();
    int		  zDimension = ((SFInt32 *) getField(11))->getValue();
    float	  zSpacing = ((SFFloat *) getField(12))->getValue();

    // first some sanity checks for the guy who has none

    if (xDimension == 0 || zDimension == 0) return;

    if (((MFFloat *) getField(3))->getSize() != xDimension * zDimension)
    {
/*
	if (_oldX != 0 && _oldX != xDimension) resizeX();
	if (_oldZ != 0 && _oldZ != zDimension) resizeZ();
	if (((MFFloat *) getField(3))->getSize() != xDimension * zDimension)
*/
	    return;
    }

    if (_normals) _normals->unref();
    if (normal == NULL) {
	generateNormals();
	_normals->ref();
	normalPerVertex = true;
    } else {
	_normals = (MFVec3f *) normal->getField(0);
	_normals->ref();
	if (normalPerVertex) {
	    if (_normals->getSize() < xDimension * zDimension * 4) {
		return;
	    }
	} else {
	    if (_normals->getSize() < (xDimension-1)* (zDimension-1)) {
		return;
	    }
	}
    }

    glFrontFace(ccw ? GL_CCW : GL_CW);
    if (solid) {
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    } else {
	glEnable(GL_CULL_FACE);
    }

    for (int j = 0; j < zDimension-1; j++) {
	for (int i = 0; i < xDimension-1; i++) {
	    glBegin(GL_TRIANGLE_STRIP);
	    glNormal3fv(_normals->getValue(i + j * (xDimension-1)));
	    glTexCoord2f((float) i / (xDimension-1), (float) j / (zDimension-1));
	    glVertex3f(i * xSpacing, HEIGHT(i, j), j * zSpacing);
	    glTexCoord2f((float) i / (xDimension-1), (float) (j+1) / (zDimension-1));
	    glVertex3f(i * xSpacing, HEIGHT(i, j+1), (j+1) * zSpacing);
	    glTexCoord2f((float) (i+1) / (xDimension-1), (float) j / (zDimension-1));
	    glVertex3f((i+1) * xSpacing, HEIGHT(i+1, j), j * zSpacing);
	    glTexCoord2f((float) (i+1) / (xDimension-1), (float) (j+1) / (zDimension-1));
	    glVertex3f((i+1) * xSpacing, HEIGHT(i+1, j+1), (j+1) * zSpacing);
 	    glEnd();
	}
    }
    glFrontFace(GL_CCW);
    glDisable(GL_CULL_FACE);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
}

void
NodeElevationGrid::generateNormals()
{
    const float  *height = ((MFFloat *) getField(3))->getValue();
    int		  xDimension = ((SFInt32 *) getField(9))->getValue();
    float	  xSpacing = ((SFFloat *) getField(10))->getValue();
    int		  zDimension = ((SFInt32 *) getField(11))->getValue();
    float	  zSpacing = ((SFFloat *) getField(12))->getValue();

    if (xDimension < 2 || zDimension < 2) return;

    Vec3f	 *faceNormals = new Vec3f[(xDimension-1) * (zDimension-1)];

    for (int i = 0; i < xDimension-1; i++) {
	for (int j = 0; j < zDimension-1; j++) {
	    Vec3f   c1((i+1) * xSpacing, HEIGHT(i+1, j), j * zSpacing);
	    Vec3f   c2(i * xSpacing, HEIGHT(i, j), j * zSpacing);
	    Vec3f   c3(i * xSpacing, HEIGHT(i, j+1), (j+1) * zSpacing);
	    Vec3f   v1 = c1 - c2;
	    Vec3f   v2 = c3 - c2;
	    Vec3f normal = v2.cross(v1);
	    normal.normalize();
	    faceNormals[i + j * (xDimension-1)] = normal;
	}
    }

    _normals = new MFVec3f(faceNormals, (xDimension-1) * (zDimension-1));

    delete faceNormals;
}

void
NodeElevationGrid::drawHandles()
{
    const float  *height = ((MFFloat *) getField(3))->getValue();
    int		  xDimension = ((SFInt32 *) getField(9))->getValue();
    float	  xSpacing = ((SFFloat *) getField(10))->getValue();
    int		  zDimension = ((SFInt32 *) getField(11))->getValue();
    float	  zSpacing = ((SFFloat *) getField(12))->getValue();

    if (xDimension < 2 || zDimension < 2) return;

    if (((MFFloat *) getField(3))->getSize() != xDimension * zDimension)
	return;

    glDisable(GL_LIGHTING);
    glPushName(0);
    glColor3f(1.0f, 1.0f, 1.0f);
    for (int j = 0; j < zDimension; j++) {
	for (int i = 0; i < xDimension; i++) {
	    glLoadName(j * xDimension + i);
	    drawHandle(Vec3f(i * xSpacing, HEIGHT(i, j), j * zSpacing));
	}
    }
    glPopName();
    glEnable(GL_LIGHTING);
}

void
NodeElevationGrid::resizeX()
{
    int		  newX = ((SFInt32 *) getField(9))->getValue();
    int		  newZ = ((SFInt32 *) getField(11))->getValue();
    const float  *oldHeight = ((MFFloat *) getField(3))->getValue();

    float     *newHeight = new float[newX * newZ];

    for (int i = 0; i < newX; i++) {
	for (int j = 0; j < _oldZ; j++) {
	    if (i < _oldX) {
		newHeight[j * newX + i] = oldHeight[j * _oldX + i];
	    } else {
		newHeight[j * newX + i] = 0.0f;
	    }
	}
    }

    _scene->execute(new FieldCommand(this, 3, new MFFloat(newHeight, newX * newZ)));
    _oldX = newX;

    delete [] newHeight;
}

void
NodeElevationGrid::resizeZ()
{
    int		  newX = ((SFInt32 *) getField(9))->getValue();
    int		  newZ = ((SFInt32 *) getField(11))->getValue();
    const float  *oldHeight = ((MFFloat *) getField(3))->getValue();

    float     *newHeight = new float[_oldX * newZ];

    for (int i = 0; i < _oldX; i++) {
	for (int j = 0; j < newZ; j++) {
	    if (j < _oldZ) {
		newHeight[j * _oldX + i] = oldHeight[j * _oldX + i];
	    } else {
		newHeight[j * _oldX + i] = 0.0f;
	    }
	}
    }

    _scene->execute(new FieldCommand(this, 3, new MFFloat(newHeight, newX * newZ)));
    _oldZ = newZ;
    
    delete [] newHeight;
}

void
NodeElevationGrid::setField(int index, FieldValue *value)
{
    if (index == 9) _oldX = ((SFInt32 *) getField(9))->getValue();
    if (index == 11) _oldZ = ((SFInt32 *) getField(11))->getValue();

    Node::setField(index, value);
}

Vec3f
NodeElevationGrid::getHandle(int handle, int *constraint, int *field)
{
    const float  *height = ((MFFloat *) getField(3))->getValue();
    int		  xDimension = ((SFInt32 *) getField(9))->getValue();
    float	  xSpacing = ((SFFloat *) getField(10))->getValue();
//    int		  zDimension = ((SFInt32 *) getField(11))->getValue();
    float	  zSpacing = ((SFFloat *) getField(12))->getValue();
    float	  x = (handle % xDimension) * xSpacing;
    float	  y = height[handle];
    float	  z = handle / xDimension * zSpacing;

    *field = 3;
    *constraint = CONSTRAIN_Y;
    return Vec3f(x, y, z);
}

void
NodeElevationGrid::setHandle(int handle, const Vec3f &v)
{
    MFFloat	*oldValue = (MFFloat *) getField(3);
    MFFloat	*newValue = new MFFloat(*oldValue);

    newValue->setValue(handle, v.y);

    _scene->setField(this, 3, newValue);
}
