/*
 * NodeImageTexture.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "stdafx.h"

#include "NodeImageTexture.h"
#include "Scene.h"
#include "Proto.h"
#include "FieldValue.h"
#include "MFString.h"
#include "SFBool.h"
#include "SFInt32.h"
#include "Image.h"
#include "Field.h"
#include "URL.h"

enum {
    IMG_STATUS_UNLOADED,
    IMG_STATUS_LOADED,
    IMG_STATUS_ERROR,
};

ProtoImageTexture::ProtoImageTexture(Scene *scene)
  : Proto(scene, "ImageTexture")
{
    addExposedField(MFSTRING, "url", new MFString(), FF_URL, NULL);
    addField(SFBOOL, "repeatS", new SFBool(true));
    addField(SFBOOL, "repeatT", new SFBool(true));
    addField(SFINT32, "mode", new SFInt32(0), 0.0f);
}

NodeImageTexture::NodeImageTexture(Scene *scene, Proto *def)
  : Node(scene, def)
{
    _image = NULL;
    _textureWidth = 0;
    _textureHeight = 0;
    _imageStatus = IMG_STATUS_UNLOADED;
    _textureName = 0;
}

NodeImageTexture::NodeImageTexture(const NodeImageTexture &node)
  : Node(node)
{
    int size = node._textureWidth * node._textureHeight * 3;
    // copy image data, if any
    if (node._image) {
	_image = new unsigned char[size];
	memcpy(_image, node._image, size);
    } else {
	_image = NULL;
    }
    _imageStatus = node._imageStatus;
    _textureWidth = node._textureWidth;
    _textureHeight = node._textureHeight;
    _textureName = 0;		    // must load its own texture though
}

NodeImageTexture::~NodeImageTexture()
{
    if (_textureName != 0) glDeleteTextures(1, &_textureName);
    delete [] _image;
}

void
NodeImageTexture::setField(int field, FieldValue *value)
{
    if (field == 0) {
	_imageStatus = IMG_STATUS_UNLOADED;
    }
    Node::setField(field, value);
}

void
NodeImageTexture::load()
{
    MFString	   *urls = (MFString *) getField(0);
    int		    width, height;
    unsigned char  *data = NULL;

    for (int i = 0; i < urls->getSize(); i++) {
	String	    path;
	URL	    url(_scene->getURL(), urls->getValue(i));
	if (urls->getValue(i).length() == 0) continue;
	_scene->Download(url, &path);

	Image	    image;
	if (image.Open(path)) {
	    width = image.GetWidth();
	    height = image.GetHeight();
	    _components = image.GetComponents();
	    data = new unsigned char [width * height * _components];
	    image.Read(data);
	    break;
	}
    }

    if (data != NULL) {
	_textureWidth = 1;
	_textureHeight = 1;

	while (_textureWidth < width)
	    _textureWidth <<= 1;

	while (_textureHeight < height)
	    _textureHeight <<= 1;

	delete [] _image;

	if (width == _textureWidth && height == _textureHeight) {
	    _image = data;
	} else {
	    _image = new unsigned char[_textureWidth * _textureHeight * 
				       _components];
	    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	    glPixelStorei(GL_PACK_ALIGNMENT, 1);
	    gluScaleImage(GL_RGB, width, height, GL_UNSIGNED_BYTE, data,
			  _textureWidth, _textureHeight, GL_UNSIGNED_BYTE, _image);
	    delete [] data;
	}
	_imageStatus = IMG_STATUS_LOADED;
    } else {
	_imageStatus = IMG_STATUS_ERROR;
    }
}

void
NodeImageTexture::bind()
{
    int	repeatS = ((SFBool *) _fields[1])->getValue();
    int	repeatT = ((SFBool *) _fields[2])->getValue();
    int mode = ((SFInt32 *) _fields[3])->getValue();

    if (_imageStatus == IMG_STATUS_UNLOADED) {
	load();
	if (_textureName != 0) glDeleteTextures(1, &_textureName);
	_textureName = 0;
    }
    if (_imageStatus == IMG_STATUS_LOADED) {
	if (_textureName == 0) {
	    glGenTextures(1, &_textureName);
	    glBindTexture(GL_TEXTURE_2D, _textureName);
	    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	    glTexImage2D(GL_TEXTURE_2D, 0, 3, _textureWidth, _textureHeight, 0,
			 GL_RGB, GL_UNSIGNED_BYTE, _image);

	} else {
	    glBindTexture(GL_TEXTURE_2D, _textureName);
	}
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, 
			repeatS ? GL_REPEAT : GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, 
			repeatT ? GL_REPEAT : GL_CLAMP);
	glEnable(GL_TEXTURE_2D);
	switch (mode) {
	  case 1:
	    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	    glEnable(GL_TEXTURE_GEN_S);
	    glEnable(GL_TEXTURE_GEN_T);
	    break;
	  case 2:
	    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	    glEnable(GL_TEXTURE_GEN_S);
	    glEnable(GL_TEXTURE_GEN_T);
	    break;
	  case 3:
	    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	    glEnable(GL_TEXTURE_GEN_S);
	    glEnable(GL_TEXTURE_GEN_T);
	    break;
	}
    }
}

void
NodeImageTexture::unbind()
{
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
}

int
NodeImageTexture::isLoaded()
{
    return _imageStatus == IMG_STATUS_LOADED;
}
