/*
 * NodeIndexedFaceSet.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <GL/gl.h>
#include "stdafx.h"

#include "NodeIndexedFaceSet.h"
#include "Proto.h"
#include "FieldValue.h"
#include "MFInt32.h"
#include "SFBool.h"
#include "SFFloat.h"
#include "SFNode.h"
#include "Node.h"
#include "MFVec3f.h"
#include "MFColor.h"
#include "Vec3f.h"


ProtoIndexedFaceSet::ProtoIndexedFaceSet(Scene *scene)
  : Proto(scene, "IndexedFaceSet")
{
    addEventIn(MFINT32, "set_colorIndex");
    addEventIn(MFINT32, "set_coordIndex");
    addEventIn(MFINT32, "set_normalIndex");
    addEventIn(MFINT32, "set_texCoordIndex");
    addExposedField(SFNODE, "color", new SFNode(NULL), NODE_COLOR);
    addExposedField(SFNODE, "coord", new SFNode(NULL), NODE_COORDINATE);
    addExposedField(SFNODE, "normal", new SFNode(NULL), NODE_NORMAL);
    addExposedField(SFNODE, "texCoord", new SFNode(NULL), NODE_TEXTURE_COORDINATE);
    addField(SFBOOL, "ccw", new SFBool(true));
    addField(MFINT32, "colorIndex", new MFInt32(), -1.0f);
    addField(SFBOOL, "colorPerVertex", new SFBool(true));
    addField(SFBOOL, "convex", new SFBool(true));
    addField(MFINT32, "coordIndex", new MFInt32(), -1.0f);
    addField(SFFLOAT, "creaseAngle", new SFFloat(0.0), 0.0f);
    addField(MFINT32, "normalIndex", new MFInt32(), -1.0f);
    addField(SFBOOL, "normalPerVertex", new SFBool(true));
    addField(SFBOOL, "solid", new SFBool(true));
    addField(MFINT32, "texCoordIndex", new MFInt32(), -1.0f);
}

NodeIndexedFaceSet::NodeIndexedFaceSet(Scene *scene, Proto *def)
  : Node(scene, def)
{
    _normals = NULL;
    _normalIndex = NULL;
    _normalsDirty = true;
}

NodeIndexedFaceSet::~NodeIndexedFaceSet()
{
    if (_normals) _normals->unref();
    if (_normalIndex) _normalIndex->unref();
}

void
NodeIndexedFaceSet::setField(int index, FieldValue *value)
{
    if (index == 1 || index == 2 || index == 7 || index == 8 
     || index == 9 || index == 10 || index == 11) {
	_normalsDirty = true;
    }

    Node::setField(index, value);
}

void
NodeIndexedFaceSet::draw()
{
    int		i;
    Node       *color = ((SFNode *) getField(0))->getValue();
    Node       *coord = ((SFNode *) getField(1))->getValue();
//    Node       *texCoord = ((SFNode *) getField(3))->getValue();
    int		ccw = ((SFBool *) getField(4))->getValue();
    MFInt32    *colorIndex = (MFInt32 *) getField(5);
    int		colorPerVertex = ((SFBool *) getField(6))->getValue();
//    int		convex = ((SFBool *) getField(7))->getValue();
    MFInt32    *coordIndex = (MFInt32 *) getField(8);
    _creaseAngle = ((SFFloat *) getField(9))->getValue();
    int		solid = ((SFBool *) getField(12))->getValue();
//    const int  *texCoordIndex = ((MFInt32 *) getField(13))->getValue();
   
    if (!coord || coord->getField(0)->getType() != MFVEC3F)
	return;

    MFVec3f		*coords = (MFVec3f *) coord->getField(0);
    MFColor		*colors;

    if (_normalsDirty) {
	getNormals(coords, coordIndex);
	_normalsDirty = false;
    }

    if (color) {
	colors = (MFColor *) color->getField(0);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
    } else {
	colors = NULL;
    }

    bool inFace = false;
    int face = 0, c;

    glFrontFace(ccw ? GL_CCW : GL_CW);
    if (solid) {
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    } else {
	glEnable(GL_CULL_FACE);
    }

    int	    numCoordIndex = coordIndex->getSize();
    int	    numNormals = _normals->getSize();
    int	    numColorIndex = colorIndex->getSize();
    int	    numColors = colors ? colors->getSize() : 0;
    int	    numCoords = coords->getSize();

    for (i = 0; i < numCoordIndex; i++) {
	int		index = coordIndex->getValue(i);
	if (index == -1) {
	    glEnd();
	    inFace = false;
	    face++;
	} else {
	    if (!inFace) {
		glBegin( GL_POLYGON );
		if (!_normalPerVertex && face < numNormals) 
		    glNormal3fv(_normals->getValue(face));
		if (color && !colorPerVertex) {
		    if (numColorIndex == 0) {
			c = face;
		    } else {
			c = colorIndex->getValue(face);
		    }
		    if (c >= 0 && c < numColors * 3)
			glColor3fv(colors->getValue(c));
		}
	    }
	    if (_normalPerVertex) {
	    	int  normal = _normalIndex->getValue(i);
    		if ( normal >= 0 && normal < numNormals)
		    glNormal3fv(_normals->getValue(normal));
	    }
	    if (color && colorPerVertex) {
		if (numColorIndex == 0) {
		    c = index;
		} else {
		    c = colorIndex->getValue(i);
		}
		if ( c >= 0 && c < numColors)
		    glColor3fv(colors->getValue(c));
	    }
	    if (index >= 0 && index < numCoords)
		glVertex3fv(coords->getValue(index));
	}
    }
    if (inFace) glEnd();

    glFrontFace(GL_CCW);
    glDisable(GL_CULL_FACE);
    glDisable(GL_COLOR_MATERIAL);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
}

void NodeIndexedFaceSet::getNormals(MFVec3f *coords, MFInt32 *coordIndex)
{
    Node       *normal = ((SFNode *) getField(2))->getValue();
    MFInt32    *normalIndex = (MFInt32 *) getField(10);

    _normalPerVertex = ((SFBool *) getField(11))->getValue();

    if (_normals) _normals->unref();
    if (_normalIndex) _normalIndex->unref();

    if (normal) {
	_normals = (MFVec3f *) normal->getField(0);
	if (normalIndex && normalIndex->getSize() > 0) {
	    _normalIndex = normalIndex;
	} else {
	    _normalIndex = coordIndex;
	}
    } else {
	generateNormals(coords, coordIndex);
	_normalPerVertex = true;
    }

    _normals->ref();
    _normalIndex->ref();
}

void
NodeIndexedFaceSet::generateNormals(MFVec3f *coords, 
				    MFInt32 *coordIndex)
{
    Array<Vec3f>	faceNormals;
    int			i;
    int			numFaces = 0;

    for (i = 0; i < coordIndex->getSize(); ) {
	Vec3f c1 = coords->getValue(coordIndex->getValue(i++));
	Vec3f c2 = coords->getValue(coordIndex->getValue(i++));
	Vec3f c3 = coords->getValue(coordIndex->getValue(i++));
	Vec3f v1 = c1 - c2;
	Vec3f v2 = c3 - c2;
	Vec3f normal = v2.cross(v1);
	normal.normalize();
	faceNormals[numFaces] = normal;
	numFaces++;
	while (i < coordIndex->getSize() && coordIndex->getValue(i) != -1) i++;
	while (i < coordIndex->getSize() && coordIndex->getValue(i) == -1) i++;
    }

    smoothNormals(coords, coordIndex, faceNormals.getData());
}

class VertexInfo {
public:
		VertexInfo(int f, bool s)
		{ face = f; smooth = s; }
    int		face;
    bool	smooth;
};

typedef List<VertexInfo *> VertexInfoList;

void
NodeIndexedFaceSet::smoothNormals(const MFVec3f *coords, 
				  const MFInt32 *coordIndex, 
				  const Vec3f *faceNormals)
{
    int				i;
    VertexInfoList::Iterator   *j;
    Array<Vec3f>		normals;
    IntArray			normalIndex;
    float			cosAngle = (float) cos(_creaseAngle);
    int				numFaces = 0;
    VertexInfoList	       *faceLists = new VertexInfoList[coords->getSize()];
    IntArray			faces;
    int				avgPos;
    int				start = 0;

    for (i = 0; i < coordIndex->getSize(); i++) {
	int v = coordIndex->getValue(i);
	normalIndex[i] = v;
	if (v == -1) {
	    faces[numFaces] = start;
	    numFaces++;
	    start = i + 1;
	} else {
    	    faceLists[v].append(new VertexInfo(numFaces, false));
	}
    }

    for (i = 0; i < coords->getSize(); i++) {
	if (!faceLists[i].first()) continue;
	int	first = faceLists[i].first()->item()->face;
	Vec3f	avg(0.0f, 0.0f, 0.0f);
	bool	flag = false;
	for (j = faceLists[i].first(); j != NULL; j = j->next()) {
	    VertexInfo	    *v = j->item();
	    if (faceNormals[first].dot(faceNormals[v->face]) > cosAngle) {
		// it's smooth, join the party
		avg += faceNormals[v->face];
		flag = true;
		v->smooth = true;
	    }
	}
	if (flag) {
	    avg.normalize();
	    avgPos = normals.size();
	    normals[avgPos] = avg;
	}
	for (j = faceLists[i].first(); j != NULL; j = j->next()) {
	    VertexInfo	    *v = j->item();
	    int index;
	    if (v->smooth) {
		index = avgPos;
	    } else {
		index = normals.size();
		normals.append(faceNormals[v->face]);
	    }

	    for (int k = faces[v->face]; ; k++ ) {
		int c = coordIndex->getValue(k);
	      if (c == -1) break;
	        if (c == i) normalIndex[k] = index;
	    }		
	    delete v;
	}
    }
    delete [] faceLists;
    _normals = new MFVec3f(normals.getData(), normals.size());
    _normalIndex = new MFInt32(normalIndex.getData(), normalIndex.size());
}
