/*
 * Proto.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include "stdafx.h"

#include "Proto.h"
#include "Scene.h"
#include "EventIn.h"
#include "EventOut.h"
#include "Field.h"
#include "ExposedField.h"
#include "Node.h"

Proto::Proto(Scene *scene, const String &name)
{
    _scene = scene;
    _name = name;
    _primaryNode = NULL;
    _nodes = NULL;
}

Proto::~Proto()
{
    int		i;

    for( i = 0; i < _fields.size(); i++) delete _fields[i];
    for( i = 0; i < _eventIns.size(); i++) delete _eventIns[i];
    for( i = 0; i < _eventOuts.size(); i++) delete _eventOuts[i];
    for( i = 0; i < _exposedFields.size(); i++) delete _exposedFields[i];

    if (_primaryNode) _primaryNode->unref();

    if (_nodes) {
	for ( i = 0; i < _nodes->size(); i++) _nodes->get(i)->unref();
    }
    delete _nodes;
}

Node *
Proto::create(Scene *scene)
{
    return new ProtoNode(scene, this);
}

void
Proto::addElement(Element *element)
{
    switch( element->getElementType() ) {
      case EL_FIELD_DEF:
        _fields.append((Field *) element);
        break;
      case EL_EVENT_IN:
        _eventIns.append((EventIn *) element);
        break;
      case EL_EVENT_OUT:
        _eventOuts.append((EventOut *) element);
        break;
      case EL_EXPOSED_FIELD:
	addExposedField((ExposedField *) element);
        break;
      default:
        _scene->errorf("internal error:  unexpected element in Proto");
        break;
    }
}

void
Proto::addField(int fieldType, const String &name, FieldValue *defaultValue,
		  float min, float max)
{
    _fields.append(new Field(fieldType, name, defaultValue, NULL,
    			     min, max, NODE_ANY));
}

void
Proto::addField(int fieldType, const String &name, FieldValue *defaultValue,
		  int nodeType)
{
    _fields.append(new Field(fieldType, name, defaultValue, NULL,
			     0.0f, 0.0f, nodeType));
}

void
Proto::addField(int fieldType, const String &name, FieldValue *defaultValue,
		  int flags, const char **strings)
{
    _fields.append(new Field(fieldType, name, defaultValue, NULL,
			     0.0f, 0.0f, NODE_ANY, flags, strings));
}

void
Proto::addEventIn(int fieldType, const String &name)
{
    _eventIns.append(new EventIn(fieldType, name));
}

void
Proto::addEventOut(int fieldType, const String &name)
{
    _eventOuts.append(new EventOut(fieldType, name));
}

void
Proto::addExposedField(int fieldType, const String &name,
		         FieldValue *defaultValue, float min,
			 float max)
{
    addExposedField(new ExposedField(fieldType, name,
   				     defaultValue, min, max));
}

void
Proto::addExposedField(int fieldType, const String &name,
		         FieldValue *defaultValue, int nodeType)
{
    addExposedField(new ExposedField(fieldType, name,
   				     defaultValue, 0.0f, 0.0f, nodeType));
}

void
Proto::addExposedField(int fieldType, const String &name,
		         FieldValue *defaultValue, int flags, const char **strings)
{
    addExposedField(new ExposedField(fieldType, name,
   				     defaultValue, 0.0f, 0.0f, 0, flags, 
				     strings));
}
void
Proto::addExposedField(ExposedField *exposedField)
{
    const String &name = exposedField->getName();
    int		 type = exposedField->getType();
    FieldValue  *value = exposedField->getValue();
    float	 min = exposedField->getMin();
    float	 max = exposedField->getMax();
    int		 flags = exposedField->getFlags();
    int		 nodeType = exposedField->getNodeType();
    const char **strings = exposedField->getStrings();

    _exposedFields.append(exposedField);

    // now add a hidden Field with the same name
    _fields.append(new Field(type, name, value, exposedField, min, max, 
			     nodeType, FF_HIDDEN | flags, strings));

    exposedField->setField(_fields.size() - 1);

    // now add a hidden EventIn called set_<name>
    char		buf[1024];
    sprintf(buf, "set_%s", (const char *) name);
    _eventIns.append(new EventIn(type, buf, exposedField));

    exposedField->setEventIn(_eventIns.size() - 1);
    // now add a hidden EventOut called <name>_changed
    sprintf(buf, "%s_changed", (const char *) name);
    _eventOuts.append(new EventOut(type, buf, exposedField));

    exposedField->setEventOut(_eventOuts.size() - 1);
}

int
Proto::lookupSimpleEventIn(const String &name) const
{
    for (int i = 0; i < _eventIns.size(); i++)
	if (_eventIns[i]->getName() == name) return i;

    return INVALID_INDEX;
}

int
Proto::lookupEventIn(const String &name) const
{
    int		index;

    if ((index = lookupSimpleEventIn(name)) == INVALID_INDEX)
    {
	// simple search failed; look for an exposedField
	if ((index = lookupExposedField(name)) != INVALID_INDEX) {

	    // now look up the corresponding eventIn
	    char		buf[1024];
	    sprintf(buf, "set_%s", (const char *) name);
	    index = lookupSimpleEventIn(buf);
	}
    }
    return index;
}

int
Proto::lookupSimpleEventOut(const String &name) const
{
    for (int i = 0; i < _eventOuts.size(); i++)
	if (_eventOuts[i]->getName() == name)
	    return i;

    return INVALID_INDEX;
}

int
Proto::lookupEventOut(const String &name) const
{
    int		index;

    if ((index = lookupSimpleEventOut(name)) == INVALID_INDEX)
    {
	// simple search failed; look for an exposedField
	if ((index = lookupExposedField(name)) != INVALID_INDEX) {

	    // now look up the corresponding eventOut
	    char		buf[1024];
	    sprintf(buf, "%s_changed", (const char *) name);
	    index = lookupSimpleEventOut(buf);
	}
    }
    return index;
}

int
Proto::lookupField(const String &name) const
{
    for (int i = 0; i < _fields.size(); i++)
	if (_fields[i]->getName() == name)
	    return i;

    return INVALID_INDEX;
}

int
Proto::lookupExposedField(const String &name) const
{
    for (int i = 0; i < _exposedFields.size(); i++)
	if (_exposedFields[i]->getName() == name)
	    return i;

    return INVALID_INDEX;
}

void
Proto::write(FILE *f, int indent) const
{
    int		i;

    for (i = 0; i < _fields.size(); i++)
	_fields[i]->write(f, indent);
    for (i = 0; i < _eventIns.size(); i++)
	_eventIns[i]->write(f, indent);
    for (i = 0; i < _eventOuts.size(); i++)
	_eventOuts[i]->write(f, indent);
    for (i = 0; i < _exposedFields.size(); i++)
	_exposedFields[i]->write(f, indent);
}

void
Proto::define(Node *primaryNode, NodeList *nodes)
{
    _primaryNode = primaryNode;
    _primaryNode->ref();
    _nodes = nodes;
    if (_nodes) {
	for ( int i = 0; i < _nodes->size(); i++) _nodes->get(i)->ref();
    }
}

ProtoNode::ProtoNode(Scene *scene, Proto *proto)
  : Node(scene, proto)
{
}
