/*
 * Scene.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "stdafx.h"

#include "swt.h"

#include "Matrix.h"
#include "Scene.h"
#include "SceneView.h"
#include "FieldValue.h"
#include "FieldCommand.h"
#include "MFieldCommand.h"
#include "CommandList.h"
#include "RouteCommand.h"
#include "UnRouteCommand.h"
#include "MoveCommand.h"
#include "Node.h"
#include "SFNode.h"
#include "SFTime.h"
#include "MFNode.h"
#include "Proto.h"
#include "parser.h"
#include "Path.h"
#include "Field.h"
#include "EventIn.h"
#include "EventOut.h"
#include "ExposedField.h"
#include "URL.h"
#include "FontInfo.h"
#include "DuneApp.h"

#include "NodeAnchor.h"
#include "NodeAppearance.h"
#include "NodeAudioClip.h"
#include "NodeBackground.h"
#include "NodeBillboard.h"
#include "NodeBox.h"
#include "NodeCollision.h"
#include "NodeColor.h"
#include "NodeColorInterpolator.h"
#include "NodeCone.h"
#include "NodeCoordinate.h"
#include "NodeCoordinateInterpolator.h"
#include "NodeCylinder.h"
#include "NodeCylinderSensor.h"
#include "NodeDirectionalLight.h"
#include "NodeElevationGrid.h"
#include "NodeExtrusion.h"
#include "NodeFog.h"
#include "NodeFontStyle.h"
#include "NodeGroup.h"
#include "NodeImageTexture.h"
#include "NodeIndexedFaceSet.h"
#include "NodeIndexedLineSet.h"
#include "NodeInline.h"
#include "NodeLOD.h"
#include "NodeMaterial.h"
#include "NodeMovieTexture.h"
#include "NodeNavigationInfo.h"
#include "NodeNormal.h"
#include "NodeNormalInterpolator.h"
#include "NodeOrientationInterpolator.h"
#include "NodePixelTexture.h"
#include "NodePlaneSensor.h"
#include "NodePointLight.h"
#include "NodePointSet.h"
#include "NodePositionInterpolator.h"
#include "NodeProximitySensor.h"
#include "NodeScalarInterpolator.h"
#include "NodeShape.h"
#include "NodeSphere.h"
#include "NodeSphereSensor.h"
#include "NodeSpotLight.h"
#include "NodeSound.h"
#include "NodeSwitch.h"
#include "NodeText.h"
#include "NodeTextureCoordinate.h"
#include "NodeTextureTransform.h"
#include "NodeTimeSensor.h"
#include "NodeTouchSensor.h"
#include "NodeTransform.h"
#include "NodeViewpoint.h"
#include "NodeVisibilitySensor.h"
#include "NodeWorldInfo.h"

#define ARRAY_SIZE(v)  ((int) (sizeof(v) / sizeof(v[0])))

#define PICK_BUFFER_SIZE	1024
#define PICK_REGION_SIZE	5

enum {
    PICKED_NODE,
    PICKED_HANDLE
};

static struct {
    const char	*name;
    int		 type;
} fieldTypes[] = {
    { "SFBool",		SFBOOL, },
    { "SFColor",	SFCOLOR, },
    { "SFFloat",	SFFLOAT, },
    { "SFImage",	SFIMAGE, },
    { "SFInt32",	SFINT32, },
    { "SFNode",		SFNODE, },
    { "SFRotation",	SFROTATION, },
    { "SFString",	SFSTRING, },
    { "SFTime",		SFTIME, },
    { "SFVec2f",	SFVEC2F, },
    { "SFVec3f",	SFVEC3F, },
    { "MFColor",	MFCOLOR, },
    { "MFFloat",	MFFLOAT, },
    { "MFInt32",	MFINT32, },
    { "MFNode",		MFNODE, },
    { "MFRotation",	MFROTATION, },
    { "MFString",	MFSTRING, },
    { "MFVec2f",	MFVEC2F, },
    { "MFVec3f",	MFVEC3F, },
};

Scene::Scene()
{
    _numSymbols = 0;

    for (int i = 0; i < ARRAY_SIZE(fieldTypes); i++) {
	_fieldTypeMap[fieldTypes[i].name] = fieldTypes[i].type;
    }

    _headlight = true;
    _numLights = 0;
    _selection = NULL;
    _selectedHandle = -1;
    _transformMode = TM_TRANSLATE;
    _unmodified = NULL;

    _protos["Anchor"] = new ProtoAnchor(this);
    _protos["Appearance"] = new ProtoAppearance(this);
    _protos["AudioClip"] = new ProtoAudioClip(this);
    _protos["Background"] = new ProtoBackground(this);
    _protos["Billboard"] = new ProtoBillboard(this);
    _protos["Box"] = new ProtoBox(this);
    _protos["Collision"] = new ProtoCollision(this);
    _protos["Color"] = new ProtoColor(this);
    _protos["ColorInterpolator"] = new ProtoColorInterpolator(this);
    _protos["Cone"] = new ProtoCone(this);
    _protos["Coordinate"] = new ProtoCoordinate(this);
    _protos["CoordinateInterpolator"] = new ProtoCoordinateInterpolator(this);
    _protos["Cylinder"] = new ProtoCylinder(this);
    _protos["CylinderSensor"] = new ProtoCylinderSensor(this);
    _protos["DirectionalLight"] = new ProtoDirectionalLight(this);
    _protos["ElevationGrid"] = new ProtoElevationGrid(this);
    _protos["Extrusion"] = new ProtoExtrusion(this);
    _protos["Fog"] = new ProtoFog(this);
    _protos["FontStyle"] = new ProtoFontStyle(this);
    _protos["Group"] = new ProtoGroup(this);
    _protos["ImageTexture"] = new ProtoImageTexture(this);
    _protos["IndexedFaceSet"] = new ProtoIndexedFaceSet(this);
    _protos["IndexedLineSet"] = new ProtoIndexedLineSet(this);
    _protos["Inline"] = new ProtoInline(this);
    _protos["LOD"] = new ProtoLOD(this);
    _protos["Material"] = new ProtoMaterial(this);
    _protos["MovieTexture"] = new ProtoMovieTexture(this);
    _protos["NavigationInfo"] = new ProtoNavigationInfo(this);
    _protos["Normal"] = new ProtoNormal(this);
    _protos["NormalInterpolator"] = new ProtoNormalInterpolator(this);
    _protos["OrientationInterpolator"] = new ProtoOrientationInterpolator(this);
    _protos["PixelTexture"] = new ProtoPixelTexture(this);
    _protos["PlaneSensor"] = new ProtoPlaneSensor(this);
    _protos["PointLight"] = new ProtoPointLight(this);
    _protos["PointSet"] = new ProtoPointSet(this);
    _protos["PositionInterpolator"] = new ProtoPositionInterpolator(this);
    _protos["ProximitySensor"] = new ProtoProximitySensor(this);
    _protos["ScalarInterpolator"] = new ProtoScalarInterpolator(this);
    _protos["Shape"] = new ProtoShape(this);
    _protos["Sound"] = new ProtoSound(this);
    _protos["Sphere"] = new ProtoSphere(this);
    _protos["SphereSensor"] = new ProtoSphereSensor(this);
    _protos["SpotLight"] = new ProtoSpotLight(this);
    _protos["Switch"] = new ProtoSwitch(this);
    _protos["Text"] = new ProtoText(this);
    _protos["TextureCoordinate"] = new ProtoTextureCoordinate(this);
    _protos["TextureTransform"] = new ProtoTextureTransform(this);
    _protos["TimeSensor"] = new ProtoTimeSensor(this);
    _protos["TouchSensor"] = new ProtoTouchSensor(this);
    _protos["Transform"] = new ProtoTransform(this);
    _protos["Viewpoint"] = new ProtoViewpoint(this);
    _protos["VisibilitySensor"] = new ProtoVisibilitySensor(this);
    _protos["WorldInfo"] = new ProtoWorldInfo(this);

    _root = createNode("Group");
    _root->ref();
    _running = false;
    _recording = false;
    _tempSave = false;

    _defaultViewpoint = (NodeViewpoint *) createNode("Viewpoint");
    _defaultViewpoint->ref();
    _currentViewpoint = _defaultViewpoint;
    _currentViewpoint->ref();
}

Scene::~Scene()
{
    int		i;

    while (!_undoStack.empty()) delete _undoStack.pop();

    while (!_redoStack.empty()) delete _redoStack.pop();

    _defaultViewpoint->unref();
    _currentViewpoint->unref();

    delete _selection;

    _root->unref();

    ProtoMap::Chain::Iterator	       *j;

    for ( i = 0; i < _protos.width(); i++) {
	for ( j = _protos.chain(i).first(); j != NULL; j = j->next()) {
	    delete j->item()->getData();
	}
    }
    for (i = 0; i < _fonts.size(); i++) {
	delete _fonts[i];
    }
}

void Scene::def(const char *nodeName, Node *value)
{
    if (value) {
	_nodeMap[nodeName] = value;
	value->setName(nodeName);
    }
}

void Scene::undef(const char *nodeName)
{
    _nodeMap[nodeName] = NULL;
}

Node *Scene::use(const char *nodeName)
{
    return _nodeMap[nodeName];
}

int
Scene::addSymbol(String s)
{
    int		&id = _symbols[s];

    if (id == 0) {
	id = _numSymbols++;
	_symbolList[id] = s;
    }
    return id;
}

const String &
Scene::getSymbol(int id) const
{
    return _symbolList[id];
}

void
Scene::setNodes(NodeList *nodes)
{
    _root->setField(0, new MFNode(nodes));
}

void
Scene::write(FILE *f, const char *url, bool tempSave)
{
    _newURL = url;
    _tempSave = tempSave;

    fprintf(f, "#VRML V2.0 utf8\n");
    _nodes.clearFlag(NODE_FLAG_TOUCHED);
    NodeList   *children = ((MFNode *) _root->getField(0))->getValue();
    for (int i = 0; i < children->size(); i++) {
	children->get(i)->write(f, 0);
    }
    if (!tempSave) {
	_unmodified = _undoStack.empty() ? (Command *) NULL : _undoStack.peek();
	_URL = url;
    }
}

void
Scene::addProto(String name, Proto *value)
{
    _protos[name] = value;
}

Proto *
Scene::getProto(String name)
{
    return _protos[name];
}

// static
bool
Scene::validRoute(Node *src, int eventOut, Node *dst, int eventIn)
{
    if (eventOut < 0 || eventOut >= src->getProto()->getNumEventOuts())
	return false;

    if (eventIn < 0 || eventIn >= dst->getProto()->getNumEventIns())
	return false;

    if (src->getProto()->getEventOut(eventOut)->getType() != 
	dst->getProto()->getEventIn(eventIn)->getType())
	return false;

    return true;
}

void
Scene::addRoute(Node *src, int eventOut, Node *dst, int eventIn)
{
    if (!validRoute(src, eventOut, dst, eventIn)) return;
    RouteUpdate		hint(src, eventOut, dst, eventIn);

    src->addOutput(eventOut, dst, eventIn);
    dst->addInput(eventIn, src, eventOut);
    UpdateViews(NULL, UPDATE_ADD_ROUTE, (Hint *) &hint);
}

void
Scene::deleteRoute(Node *src, int eventOut, Node *dst, int eventIn)
{
    if (eventOut < 0 || eventOut >= src->getProto()->getNumEventOuts())
	return;

    if (eventIn < 0 || eventIn >= dst->getProto()->getNumEventIns())
	return;

    RouteUpdate		hint(src, eventOut, dst, eventIn);

    src->removeOutput(eventOut, dst, eventIn);
    dst->removeInput(eventIn, src, eventOut);
    UpdateViews(NULL, UPDATE_DELETE_ROUTE, (Hint *) &hint);
}

int
Scene::getFieldType(String name)
{
    return _fieldTypeMap[name];
}

const char *
Scene::getFieldTypeName(int type)
{
    for (int i = 0; i < ARRAY_SIZE(fieldTypes); i++)
	if (fieldTypes[i].type == type)
	    return fieldTypes[i].name;
    return "";
}

void
Scene::errorf(const char *fmt, ...)
{
    va_list	ap;
    char	buf[1024], buf2[1024];

    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    sprintf(buf2, "%d:  %s", lineno, buf);
    _compileErrors += buf;
}

void
Scene::invalidNode(const char *name)
{
    errorf("invalid DEF name \"%s\"\n", name);
}

void 
Scene::invalidField(const char *node, const char *field)
{
    errorf("node \"%s\" has no field \"%s\", ignored", node, field);
}

const char *
Scene::parse(FILE *f)
{
    inputFile = f;
    scene = this;
    lineno = 1;
    _compileErrors = "";
    yyparse();
    return _compileErrors;
}

void
Scene::add(Command *cmd)
{
    _undoStack.push(cmd);

    while (!_redoStack.empty()) {
	delete _redoStack.pop();
    }
}

void
Scene::execute(Command *cmd)
{
    cmd->execute();
    add(cmd);
}

void
Scene::backupField(Node *node, int field)
{
    CommandList	    *list = new CommandList();

    if (isRecording()) {
	Interpolator	    *interp = findUpstreamInterpolator(node, field);

	if (interp) {
	    interp->backup(list);
	}
    }
    list->append(new FieldCommand(node, field));
    add(list);
}

Interpolator *
Scene::findUpstreamInterpolator(Node *node, int field) const
{
    // is this field an ExposedField?
    ExposedField    *e = node->getProto()->getField(field)->getExposedField();

    if (e) {
	const SocketList::Iterator	*i;

	// check for interpolator routed to the corresponding EventIn;

	for (i = node->getInput(e->getEventIn()).first(); i != NULL; i = i->next()) {
	    if (i->item()._node->getNodeClass() == NODE_INTERPOLATOR)
		return ((Interpolator *) i->item()._node);
	}
    }
    return NULL;
}

void
Scene::setField(Node *node, int field, FieldValue *value)
{
    if (isRecording()) {
	Interpolator	    *interp = findUpstreamInterpolator(node, field);

	if (interp) {
	    interp->recordKey(value);
	}
    }
    node->setField(field, value);
    OnFieldChange(node, field);
}

void
Scene::undo()
{
    if (_undoStack.empty()) {
	assert(false);
	return;
    }

    Command	*change = _undoStack.pop();
    change->undo();
    _redoStack.push(change);
}

void
Scene::redo()
{
    if (_redoStack.empty()) {
	assert(false);
	return;
    }

    Command	*change = _redoStack.pop();
    change->execute();
    _undoStack.push(change);
}

void
Scene::drawScene(bool pick, int x, int y)
{
    GLint	v[4];
    float	aspect;
    int		i;

    glGetIntegerv(GL_VIEWPORT, v);

    if (v[3])
	aspect = (GLfloat)v[2]/v[3];
    else	// don't divide by zero, not that we should ever run into that...
	aspect = 1.0f;

    _numLights = 0;
    _headlight = true;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    if (pick) gluPickMatrix(x, y, PICK_REGION_SIZE, PICK_REGION_SIZE, v);
    gluPerspective(45.0, aspect, 1.0f, 7000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glClearColor(0.0F, 0.0F, 0.0F, 1.0F);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glEnable(GL_LIGHTING);
    glEnable(GL_NORMALIZE);
    glEnable( GL_DEPTH_TEST );
    glDepthFunc( GL_LEQUAL );

    // no scene ambient light, please
    GLint    zero[4] = { 0, 0, 0, 0 };
    glLightModeliv(GL_LIGHT_MODEL_AMBIENT, zero);

    // first pass:  pre-draw traversal
    // enable PointLights and SpotLights; pick up ViewPoints, Fogs.
    // Backgrounds and TimeSensors

    glPushMatrix();
    applyCamera();
    _viewpoints.resize(0);
    _fogs.resize(0);
    _backgrounds.resize(0);
    _timeSensors.resize(0);

    _root->preDraw();

    if (_fogs.size() > 0) ((NodeFog *) _fogs[0])->apply();
    glPopMatrix();

    // second pass:  main drawing traversal

    glPushMatrix();
    if (_headlight) enableHeadlight();
    glPushName(PICKED_NODE);
    if (_backgrounds.size() > 0) {
	glPushName(0);		    // FIXME:  can't pick backgrounds, yet
	((NodeBackground *) _backgrounds[0])->apply();
	glPopName();
    }
    applyCamera();
    _root->draw();

    for (i = 0; i < _numLights; i++) {
	glDisable((GLenum) (GL_LIGHT0 + i));
    }

    _numLights = 0;
    glDisable(GL_FOG);
    glPopMatrix();

    // post-draw phase
    // draw handles for all nodes in the current selection path

    glPushMatrix();
    enableHeadlight();
    applyCamera();

    glLoadName(PICKED_HANDLE);
    drawHandles();
    glPopName();

    glDisable(GL_LIGHT0);
    glPopMatrix();
}

Path *
Scene::pick(int x, int y)
{
    GLuint	pickBuffer[PICK_BUFFER_SIZE];

    glSelectBuffer(PICK_BUFFER_SIZE, pickBuffer);
    glRenderMode(GL_SELECT);
    glInitNames();

    drawScene(true, x, y);

    int hits = glRenderMode(GL_RENDER);
    if (hits) {
	return processHits(hits, pickBuffer);
    } else {
	return new Path(NULL, 0, this);
    }
}

Path *
Scene::processHits(unsigned hits, GLuint *pickBuffer)
{
    GLuint     *path = NULL;
    int		pathLen;
    GLuint	depth = UINT_MAX;

    _selectedHandle = -1;

    for (unsigned i = 0; i < hits; i++) {
	unsigned    numNames = *pickBuffer++;
	unsigned    minDepth = *pickBuffer++;
	/* unsigned    maxDepth = * */ pickBuffer++;
	if (*pickBuffer == PICKED_NODE) {
	    if (_selectedHandle == -1 || minDepth < depth) {
		path = pickBuffer + 1;
		pathLen = numNames - 1;
		depth = minDepth;
	    }
	} else if (*pickBuffer == PICKED_HANDLE) {
	    path = pickBuffer + 1;
	    pathLen = numNames - 2;
	    _selectedHandle = pickBuffer[numNames-1];
	    depth = minDepth;
	} else {
	    assert(false);
	}
	pickBuffer += numNames;
    }
    assert( path != NULL );
    return new Path((int *) path, pathLen, this);
}

void
Scene::transform(const Path *path)
{
    assert(path != NULL);
    applyCamera();
    const NodeList    *nodes = path->getNodes();
    int		size = nodes->size() - 1;
    for (int i = 0; i < size; i++) {
	nodes->get(i)->transform();
    }
}

void
Scene::drawHandles()
{
    glDisable(GL_DEPTH_TEST);
    switch (TheApp->GetHandleMode()) {
      case HM_NONE:
	break;
      case HM_SELECTED:
	if (_selection != NULL) {
	    glPushMatrix();
	    int	len = _selection->getPathLen();
	    const int    *path = _selection->getPath();
	    Node *node = _root;
	    for (int i = 0; i < len;) {
		int field = path[i++];
		glPushName(field);
    		FieldValue  *value = node->getField(field);
		if (i >= len) {
		    break;
		}
		glPushName(path[i]);
		if (value->getType() == SFNODE) {
		    node = ((SFNode *) value)->getValue(); i++;
		} else if (value->getType() == MFNODE) {
		    node = ((MFNode *) value)->getValue(path[i++]);
		} else {
		    assert(false);
		}
		node->drawHandles();
		node->transform();
	    }
	    glPopMatrix();
	}
	break;
      case HM_ALL:
	drawHandlesRec(_root);
	break;
    }
}

void
Scene::drawHandlesRec(Node *node) const
{
    int	    numFields = node->getProto()->getNumFields();

    glPushMatrix();
    node->drawHandles();
    node->transform();
    for (int i = 0; i < numFields; i++) {
    	FieldValue  *value = node->getField(i);
	if (value->getType() == SFNODE) {
	    Node	*child = ((SFNode *) value)->getValue();
	    if (child) {
		glPushName(i);
		glPushName(0);
		drawHandlesRec(child);
		glPopName();
		glPopName();
	    }
	} else if (value->getType() == MFNODE) {
	    glPushName(i);
	    glPushName(0);
	    MFNode  *v = (MFNode *) value;
	    int	     n = v->getSize();
	    for (int j = 0; j < n; j++) {
		glLoadName(j);
		drawHandlesRec(v->getValue(j));
	    }
	    glPopName();
	    glPopName();
	}
    }
    glPopMatrix();
}

void
Scene::enableHeadlight()
{
    GLenum light = (GLenum) allocateLight();
    static float	pos[4] = {0.0f, 0.0f, 1.0f, 0.0f};
    static float	ambientColor[4] = {0.0f, 0.0f, 0.0f, 1.0f};
    static float	diffuseColor[4] = {1.0f, 1.0f, 1.0f, 1.0f};

    glLightfv(light, GL_AMBIENT, ambientColor);
    glLightfv(light, GL_DIFFUSE, diffuseColor);
    glLightfv(light, GL_POSITION, pos);
    glLightfv(light, GL_SPECULAR, diffuseColor);
    glLightf(light, GL_SPOT_CUTOFF, 180.0f);
    glLightf(light, GL_SPOT_EXPONENT, 0.0f);
    glLightf(light, GL_CONSTANT_ATTENUATION, 1.0f);
    glLightf(light, GL_LINEAR_ATTENUATION, 0.0f);
    glLightf(light, GL_QUADRATIC_ATTENUATION, 0.0f);
    glEnable(light);
}

// allocateLight()
//
// reserve an openGL light

int
Scene::allocateLight()
{
    GLint		maxLights;
    glGetIntegerv(GL_MAX_LIGHTS, &maxLights);

    if (_numLights >= maxLights) {
	errorf("too many lights!");
	return GL_LIGHT0;
    }
    
    return (GL_LIGHT0 + _numLights++);
}

int
Scene::freeLight()
{
    return GL_LIGHT0 + --_numLights;
}

void
Scene::projectPoint(float x, float y, float z, float *wx, float *wy, float *wz)
{
    GLdouble	mmat[16], pmat[16];
    GLdouble	winx, winy, winz;
    GLint	viewport[4];

    glGetDoublev(GL_MODELVIEW_MATRIX, mmat);
    glGetDoublev(GL_PROJECTION_MATRIX, pmat);
    glGetIntegerv(GL_VIEWPORT, viewport);

    gluProject(x, y, z, mmat, pmat, viewport, &winx, &winy, &winz);

    *wx = (float) winx;
    *wy = (float) winy;
    *wz = (float) winz;
}

void
Scene::unProjectPoint(float wx, float wy, float wz, float *x, float *y, float *z)
{
    GLdouble	mmat[16], pmat[16];
    GLdouble	objx, objy, objz;
    GLint	viewport[4];

    glGetDoublev(GL_MODELVIEW_MATRIX, mmat);
    glGetDoublev(GL_PROJECTION_MATRIX, pmat);
    glGetIntegerv(GL_VIEWPORT, viewport);

    gluUnProject(wx, wy, wz, mmat, pmat, viewport, &objx, &objy, &objz);

    *x = (float) objx;
    *y = (float) objy;
    *z = (float) objz;
}

void
Scene::addViewpoint(Node *viewpoint)
{
    _viewpoints.append(viewpoint);
}

void
Scene::addFog(Node *fog)
{
    _fogs.append(fog);
}

void
Scene::addBackground(Node *background)
{
    _backgrounds.append(background);
}

void
Scene::addTimeSensor(Node *timeSensor)
{
    _timeSensors.append(timeSensor);
}

void
Scene::moveCamera(float dx, float dy, float dz)
{
    Vec3f	pos = _currentViewpoint->getPosition();
    Quaternion	rot = _currentViewpoint->getOrientation();

    if (!isRecording() && _currentViewpoint != _defaultViewpoint) {
	_currentViewpoint = _defaultViewpoint;
	_currentViewpoint->setOrientation(rot);
    }
    _currentViewpoint->setPosition(pos + rot * Vec3f(dx, dy, dz));
}

void
Scene::turnCamera(float x, float y, float z, float ang)
{
    Quaternion	    rot = _currentViewpoint->getOrientation();
    Quaternion	    r(Vec3f(x, y, z), ang);

    _currentViewpoint->setOrientation(r * rot);
}

void
Scene::orbitCamera(float dtheta, float dphi)
{
    if (!isRecording() && _currentViewpoint != _defaultViewpoint) {
	_currentViewpoint = _defaultViewpoint;
    }

    Vec3f	pos(_currentViewpoint->getPosition());
    Quaternion	rot(_currentViewpoint->getOrientation());
    Quaternion	up(Vec3f(0.0f, 1.0f, 0.0f), dtheta);
    Quaternion	around(Vec3f(1.0f, 0.0f, 0.0f), dphi);
    Quaternion	newRot(up * around * rot);
    newRot.normalize();

    Vec3f   newPos(rot.conj() * newRot * pos);

    _currentViewpoint->setPosition(newPos);
    _currentViewpoint->setOrientation(newRot);
}

Node *
Scene::createNode(const char *nodeType)
{
    Proto *def = _protos[nodeType];

    return def ? def->create(this) : (Node *) NULL;
}

void
Scene::addNode(Node *node)
{
    _nodes.append(node);
}

String
Scene::generateUniqueNodeName(Node *node)
{
    char	buf[512];
    const char *name = node->getProto()->getName();

    for (int i = 1; ; i++) {
	sprintf(buf, "%s%d", name, i);
	if (use(buf) == NULL) break;
    }
    def(buf, node);
    return buf;
}

void
Scene::removeNode(Node *node)
{
    _nodes.remove(_nodes.find(node));
}

void
Scene::setSelection(const Path *path)
{
    if (_selection != path) {
	delete _selection;
	_selection = path;
	_nodes.clearFlag(NODE_FLAG_SELECTED);
	_selection->getNodes()->setFlag(NODE_FLAG_SELECTED);
    }
}

void
Scene::setSelection(Node *node)
{
    int		len = 0;
    Node       *n;

    for (n = node; n->getNumParents() > 0; n = n->getParent(0)) {
	len += 2;
    }

    if (len > 0) {
	int	   *list = new int[len];

	int i = len-1;
    
	for (n = node; n->getNumParents() > 0; n = n->getParent(0)) {
	    Node	   *parent = n->getParent(0);
	    int	    field = n->getParentField(0);
	    list[i--] = parent->findChild(n, field);
	    list[i--] = field;
	}

	setSelection(new Path(list, len, this));
	delete [] list;
    } else {
	// select root node
	setSelection(new Path(NULL, 0, this));
    }
}

bool
Scene::isModified() const
{
    if (_undoStack.empty()) {
	return _unmodified != NULL;
    } else {
	return _unmodified != _undoStack.peek();
    }
}

void
Scene::applyCamera()
{
    _currentViewpoint->apply();
}

void
Scene::start()
{
    _running = true;
    if (_viewpoints.size() > 0) {
	_currentViewpoint = (NodeViewpoint *) _viewpoints[0];
    } else {
	_currentViewpoint = _defaultViewpoint;
    }
    double	t = swGetCurrentTime();

    for (int i = 0; i < _timeSensors.size(); i++) {
	((NodeTimeSensor *) _timeSensors[i])->start(t);
    }
}

void
Scene::stop()
{
    _running = false;
}

void
Scene::updateTime()
{
    double	t = swGetCurrentTime();

    for (int i = 0; i < _timeSensors.size(); i++) {
	((NodeTimeSensor *) _timeSensors[i])->setTime(t);
    }
    UpdateViews(NULL, UPDATE_TIME);
}

NodeViewpoint *
Scene::getCamera() const
{
    return _currentViewpoint;
}

void
Scene::AddView(SceneView *view)
{
    _views.append(view);
}

void
Scene::RemoveView(SceneView *view)
{
    _views.remove(_views.find(view));
}

void
Scene::OnFieldChange(Node *node, int field, int index)
{
    FieldUpdate         hint(node, field, index);

    UpdateViews(NULL, UPDATE_FIELD, (Hint *) &hint);
}

void
Scene::OnAddNode(Node *node, Node *dest, int field)
{
    NodeUpdate		hint(node, dest, field);

    UpdateViews(NULL, UPDATE_ADD_NODE, (Hint *) &hint);
}

void
Scene::OnRemoveNode(Node *node, Node *src, int field)
{
    NodeUpdate		hint(node, src, field);

    UpdateViews(NULL, UPDATE_REMOVE_NODE, (Hint *) &hint);
}

void
Scene::UpdateViews(SceneView *sender, int type, Hint *hint)
{
    for (List<SceneView *>::Iterator *i = _views.first(); i != NULL; i = i->next())
    {
	SceneView	*view = i->item();
	if (view != sender)
	{
	    view->OnUpdate(sender, type, hint);
	}
    }
}

static void
BackupRoutesRec(Node *node, CommandList *list)
{
    int                         i;
    SocketList::Iterator       *j;

    if (!node) return;

    if (node->getNumParents() > 1) return;

    for (i = 0; i < node->getProto()->getNumEventIns(); i++) {
	for (j = node->getInput(i).first(); j != NULL; j = j->next()) {
	    const Socket            &s = j->item();
	    list->append(new UnRouteCommand(s._node, s._index,
					    node, i));
	}
    }
    for (i = 0; i < node->getProto()->getNumEventOuts(); i++) {
	for (j = node->getOutput(i).first(); j != NULL; j = j->next()) {
	    const Socket	    &s = j->item();
	    list->append(new UnRouteCommand(node, i, 
					    s._node, s._index));
	}
    }

    for (i = 0; i < node->getProto()->getNumFields(); i++) {
	FieldValue  *v = node->getField(i);
	if (v->getType() == SFNODE) {
	    BackupRoutesRec(((SFNode *) v)->getValue(), list);
	} else if (v->getType() == MFNODE) {
	    int         size = ((MFNode *) v)->getSize();
	    for (int k = 0; k < size; k++) {
		BackupRoutesRec(((MFNode *) v)->getValue(k), list);
	    }
	}
    }
}

void
Scene::DeleteSelected()
{
    if (_selection && _selection->getNode() != _root) {
	Node   *node = _selection->getNode();
	int     field = _selection->getField();
	Node   *parent = _selection->getParent();
	int     parentField = _selection->getParentField();

	if (field == -1 && parentField != -1) {
	    CommandList *list = new CommandList();
	    if (node->getNumParents() == 1) {
		BackupRoutesRec(node, list);
	    }
	    list->append(new MoveCommand(node, parent, parentField, NULL, -1));
	    execute(list);
	}
    }
}

int
Scene::OnDragOver(Node *src, Node *srcParent, int srcField, 
		  Node *dest, int destField, int modifiers)
{
    int	rc = 0;

    if (src && dest) {
	if (destField == -1) destField = dest->findValidField(src);
	if (dest->validChildType(destField, src->getNodeClass())) {
	    if ((modifiers & SW_CONTROL) && (modifiers & SW_SHIFT)
		&& dest != src && !dest->hasAncestor(src)) {
		rc = SW_DRAG_LINK;
	    } else if (modifiers & SW_CONTROL) {
		rc = SW_DRAG_COPY;
	    } else if (dest != src
		   && !dest->hasAncestor(src)
		   && dest->findChild(src, destField) == -1) {
		rc = SW_DRAG_MOVE;
	    }
	}
    }
    return rc;
}

int
Scene::OnDrop(Node *src, Node *srcParent, int srcField, 
	      Node *dest, int destField, int modifiers)
{
    int effect = OnDragOver(src, srcParent, srcField, dest, destField, modifiers);
    if (src && dest) {
	if (destField == -1) destField = dest->findValidField(src);
	switch(effect) {
	  case SW_DRAG_COPY:
	    execute(new MoveCommand(src->copy(), NULL, -1, dest, destField));
	    break;
	  case SW_DRAG_MOVE:
	    execute(new MoveCommand(src, srcParent, srcField, dest, destField));
	    break;
	  case SW_DRAG_LINK:
	    execute(new MoveCommand(src, NULL, -1, dest, destField));
	    break;
	}
	return 1;
    } else {
	return 0;
    }
}

bool
Scene::Download(const URL &url, String *path)
{
    *path = url.ToPath();
    return true;
}

FontInfo *
Scene::LoadGLFont(const char *fontName, const char *style)
{
    int		i;
    int		styleId;

    // handle "special" font names
    if (!strcmp(fontName, "SERIF")) {
	fontName = "Times New Roman";
    } else if (!strcmp(fontName, "SANS")) {
	fontName = "Arial";
    } else if (!strcmp(fontName, "TYPEWRITER")) {
	fontName = "Courier New";
    }
    
    if (!strcmp(style, "BOLD")) {
	styleId = SW_BOLD;
    } else if (!strcmp(style, "ITALIC")) {
	styleId = SW_ITALIC;
    } else if (!strcmp(style, "BOLDITALIC")) {
	styleId = SW_BOLD | SW_ITALIC;
    } else {
	styleId = SW_PLAIN;
    }

    // look for font in cache

    for (i = 0; i < _fonts.size(); i++) {
	if (!strcmp(_fonts[i]->name, fontName) && _fonts[i]->style == styleId) {
	    return _fonts[i];
	}
    }

    // create some font outlines

    FontInfo        *info = new FontInfo();

    info->displayListBase =
    	swLoadGLFont(fontName, styleId, info->kernX, info->kernY);

    if (info->displayListBase == 0) {
	delete info;
	return NULL;
    } else {
	info->name = fontName;
	info->style = styleId;
	_fonts.append(info);
	return info;
    }
}
