/*
 * Scene3DView.cpp
 *
 * Copyright (C) 1999 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <GL/glu.h>

#include "stdafx.h"

#include "swt.h"

#include "Scene3DView.h"
#include "Scene.h"
#include "Path.h"
#include "Node.h"
#include "Matrix.h"
#include "SFRotation.h"
#include "SFVec3f.h"
#include "EulerAngles.h"
#include "NodeViewpoint.h"

Scene3DView::Scene3DView(Scene *scene, SWND parent)
  : SceneView(scene, parent)
{
    _dc = NULL;
    _glc = NULL;
    _trackMouse = false;
    _cursorArrow = swLoadCursor(SW_CURSOR_ARROW);
    _cursorZoom = swLoadCursor(SW_CURSOR_DBL_ARROW_VERT);
    _cursorRotate = swLoadCursor(SW_CURSOR_DBL_ARROW_VERT);
}

Scene3DView::~Scene3DView()
{
    swDestroyDC(_dc);
    swDestroyGLContext(_glc);
    swDestroyCursor(_cursorArrow);
    swDestroyCursor(_cursorZoom);
    swDestroyCursor(_cursorRotate);
}

void Scene3DView::OnUpdate(SceneView *sender, int type, Hint *hint)
{
    swInvalidateWindow(_wnd);
}

void Scene3DView::OnDraw(int /* x */, int /* y */,
			 int /* width */, int /* height */) 
{
    int		width, height;
    if (!_dc) _dc = swCreateDC(_wnd);
    if (!_glc) _glc = swCreateGLContext(_dc);
    swMakeCurrent(_dc, _glc);
    swGetSize(_wnd, &width, &height);
    glViewport(0, 0, width, height);
    _scene->drawScene();
    swSwapBuffers(_dc, _glc);
}

void Scene3DView::OnSize(int width, int height) 
{
    swInvalidateWindow(_wnd);
}

void Scene3DView::OnKeyDown(int key, int x, int y, int modifiers) 
{
    switch (key) {
      case SW_UP:
	_scene->moveCamera(0.0f, 0.0f, -1.0f);
	swInvalidateWindow(_wnd);
	break;
      case SW_DOWN:
	_scene->moveCamera(0.0f, 0.0f, 1.0f);
	swInvalidateWindow(_wnd);
	break;
      case SW_KEY_LEFT:
	_scene->turnCamera(0.0f, 1.0f, 0.0f, 0.10f);
	swInvalidateWindow(_wnd);
	break;
      case SW_KEY_RIGHT:
	_scene->turnCamera(0.0f, 1.0f, 0.0f, -0.10f);
	swInvalidateWindow(_wnd);
	break;
      default:
	SceneView::OnKeyDown(key, x, y, modifiers);
	break;
    }
}

void Scene3DView::OnLButtonDown(int x, int y, int modifiers) 
{
    Path   *path;
    _backedUp = false;
    int		width, height;
    swGetSize(_wnd, &width, &height);

    swMakeCurrent(_dc, _glc);
    swSetFocus(_wnd);

    if (modifiers & SW_CONTROL) {
	swSetCursor(_wnd, _cursorRotate);
    } else if (modifiers & SW_SHIFT) {
	swSetCursor(_wnd, _cursorZoom);
    } else {
	path = _scene->pick(x, height - y);
	if (path) {
	    _scene->setSelection(path);
	    _scene->UpdateViews(NULL, UPDATE_SELECTION);
	    Node   *node = _scene->getSelection()->getNode();
	    int	     handle = _scene->getSelectedHandle();

	    if (handle != -1) {
		int	    constraint = CONSTRAIN_NONE, field;
		Vec3f	v = node->getHandle(handle, &constraint, &field);

		_backedUp = false;

		// now calculate an offset in object space from the mouse 
		// to the actual handle position

		glPushMatrix();
		Vec3f	    o;
		float	    hx, hy, hz;
		_scene->transform(_scene->getSelection());
		node->transformForHandle(handle);
		_scene->projectPoint(v.x, v.y, v.z, &hx, &hy, &hz);
		_scene->unProjectPoint((float) x, (float) height - y, hz,
				       &o.x, &o.y, &o.z);
		glPopMatrix();

		_offset = v - o;
	    }
	}
    }
    _trackMouse = true;
    swSetCapture(_wnd);
    _trackX = x;
    _trackY = y;
}

void Scene3DView::OnMouseMove(int x, int y, int modifiers) 
{
    int	    handle = _scene->getSelectedHandle();
    int	    width, height;

    swMakeCurrent(_dc, _glc);
    swGetSize(_wnd, &width, &height);

    if (_trackMouse) {
	float dx = (float) x - _trackX;
	float dy = (float) y - _trackY;
	if (modifiers & SW_CONTROL) {
	    _scene->orbitCamera(-DEG2RAD(dx * 0.5f), -DEG2RAD(dy * 0.5f));
	    _trackX = x;
	    _trackY = y;
    	    swInvalidateWindow(_wnd);
	} else if (modifiers & SW_SHIFT) {
	    _scene->moveCamera(0.0f, 0.0f, dy * 0.25f);
	    _trackX = x;
	    _trackY = y;
    	    swInvalidateWindow(_wnd);
	} else if (handle != -1) {
	    Node   *node = _scene->getSelection()->getNode();
	    int		constraint = CONSTRAIN_NONE, field;
	    Vec3f	old = node->getHandle(handle, &constraint, &field);
	    Vec3f	v = old;
	    if (constraint == CONSTRAIN_NONE) {
		float hx, hy, hz;
		glPushMatrix();
		glLoadIdentity();
		_scene->transform(_scene->getSelection());
		node->transformForHandle(handle);
		_scene->projectPoint(old.x, old.y, old.z, &hx, &hy, &hz);
		_scene->unProjectPoint((float) x, (float) height - y, hz, &v.x, &v.y, &v.z);
		glPopMatrix();
		v += _offset;
	    } else {
		float x1, y1, z1;
		float x2, y2, z2;
		Matrix mat;
		glPushMatrix();
		glLoadIdentity();
		_scene->transform(_scene->getSelection());
		glGetFloatv(GL_MODELVIEW_MATRIX, mat);
		node->transformForHandle(handle);
		_scene->unProjectPoint((float) x, (float) height - y, 0.0f, &x1, &y1, &z1);
		_scene->unProjectPoint((float) x, (float) height - y, 1.0f, &x2, &y2, &z2);
		glPopMatrix();

		x1 += _offset.x;
		y1 += _offset.y;
		z1 += _offset.z;
		x2 += _offset.x;
		y2 += _offset.y;
		z2 += _offset.z;

		switch (constraint) {
		  case CONSTRAIN_X:
		    y1 -= v.y;
		    z1 -= v.z;
		    y2 -= v.y;
		    z2 -= v.z;
		    v.x = constrainLine(y1, z1, x1, y2, z2, x2);
		    break;
		  case CONSTRAIN_Y:
		    x1 -= v.x;
		    z1 -= v.z;
		    x2 -= v.x;
		    z2 -= v.z;
		    v.y = constrainLine(x1, z1, y1, x2, z2, y2);
		    break;
		  case CONSTRAIN_Z:
		    x1 -= v.x;
		    y1 -= v.y;
		    x2 -= v.x;
		    y2 -= v.y;
		    v.z = constrainLine(x1, y1, z1, x2, y2, z2);
		    break;
		  case CONSTRAIN_XY:
		    constrainPlane(x1, y1, z1, x2, y2, z2, &v.x, &v.y);
		    v.z = 0.0f;
		    break;
		  case CONSTRAIN_SPHERE:
		    v = constrainSphere(v, dx, dy, mat);
		    _trackX = x;
		    _trackY = y;
		    break;
		}
	    }
	    if (v != old) {
		if (!_backedUp) {
		    _scene->backupField(node, field);
		    _backedUp = true;
		}
		node->setHandle( handle, v);
	    }
	}
    }
}

void Scene3DView::OnLButtonUp(int x, int y, int modifiers) 
{
    _trackMouse = false;
    swSetCursor(_wnd, _cursorArrow);
    swReleaseCapture(_wnd);
    _scene->setSelectedHandle(-1);
}

float Scene3DView::constrainLine(float x1, float y1, float z1, float x2, float y2, float z2) const
{
    float	dx = x1 - x2;
    float	dy = y1 - y2;
    if (dx == 0.0f && dy == 0.0f) 
	return z2;
    float	alpha = (x1 * dx + y1 * dy) / (dx * dx + dy * dy);

    return (1.0f - alpha) * z1 + alpha * z2;
}

void Scene3DView::constrainPlane(float x1, float y1, float z1, float x2, float y2, float z2, float *x, float *y) const
{
    if(EQUALF(z1, z2)) {
	*x = *y = 0.0f;
	return;
    }

    float	alpha = -z1 / (z2 - z1);
    
    *x = (1.0f - alpha) * x1 + alpha * x2;
    *y = (1.0f - alpha) * y1 + alpha * y2;
}

Vec3f Scene3DView::constrainSphere(const Vec3f &euler, float dx, float dy,
				   const Matrix &mat)
{
    SFRotation	sfr(euler, 0);
    Quaternion	cam = _scene->getCamera()->getOrientation();

    Quaternion  ry(Vec3f(0.0f, 1.0f, 0.0f), DEG2RAD(dx));
    Quaternion	rx(Vec3f(1.0f, 0.0f, 0.0f), DEG2RAD(dy));
    Quaternion  r = sfr.getQuat() * cam.conj() * ry * rx * cam;
    return SFRotation(r).getEulerAngles(0);
}
