/*
 * browser.c - unix interface to web browser
 *
 * Copyright (C) 2000 Stephen F. White
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file "COPYING" for details); if 
 * not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 * Cambridge, MA 02139, USA.
 */

#include <X11/Intrinsic.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>

#include "swt.h"

typedef struct SBrowser {
    STABLE			prefs;
    char		       *command;
    int				useRemote;
    char		       *remoteCommand;
    pid_t			pid;
} SBrowser;

extern XtAppContext TheAppContext;

static int	browserLaunch(SBROWSER browser, const char *path, SWND wnd);
static int	browserRemote(SBROWSER browser, const char *path);
static void	childInput(XtPointer closure , int *source, XtInputId *id);
static int	waitForProcess(pid_t pid);
static pid_t	createProcess(const char *cmdline, SWND wnd);
static pid_t	createSimpleProcess(const char *cmdline);

/*******************/
/* browser preview */
/*******************/

extern SBROWSER
swBrowserInit(STABLE prefs)
{
    SBrowser	*browser = malloc(sizeof(SBrowser));
    browser->prefs = prefs;
    browser->command = strdup(swGetPreference(prefs, "PreviewCommand",
    					      "netscape %s"));
    browser->useRemote = swGetIntPreference(prefs, "PreviewUseRemote", TRUE);
    browser->remoteCommand = strdup(swGetPreference(prefs,
					"PreviewRemoteCommand",
				        "netscape -remote OpenURL\\(%s\\)"));
    browser->pid = 0;

    return browser;
}

extern void
swBrowserGetSettings(SBROWSER browser, const char **command, int *useRemote,
		     const char **remoteCommand, const char **application,
		     const char **topic)
{
    *command = browser->command;
    *useRemote = browser->useRemote;
    *remoteCommand = browser->remoteCommand;
    *application = NULL;
    *topic = NULL;
}

extern void
swBrowserSetSettings(SBROWSER browser, const char *command, int useRemote,
		     const char *remoteCommand, const char *application,
		     const char *topic)
{
    free(browser->command);
    free(browser->remoteCommand);

    browser->command = strdup(command);
    browser->useRemote = useRemote;
    browser->remoteCommand = strdup(remoteCommand);

    swSetPreference(browser->prefs, "PreviewCommand", browser->command);
    swSetIntPreference(browser->prefs, "PreviewUseRemote", browser->useRemote);
    swSetPreference(browser->prefs, "PreviewRemoteCommand", browser->remoteCommand);
}

extern void
swBrowserPreview(SBROWSER browser, const char *path, SWND wnd)
{
    if (browser->useRemote) {
        if (!browserRemote(browser, path)) {
            browserLaunch(browser, path, wnd);
        }
    } else {
        browserLaunch(browser, path, wnd);
    }
}

extern void
swBrowserShutdown(SBROWSER browser)
{
    if (!browser) return;

    if (browser->pid != 0) {
	kill(browser->pid, SIGTERM);
	waitForProcess(browser->pid);
    }
    free(browser);
}

static int
browserLaunch(SBROWSER browser, const char *path, SWND wnd)
{
    char        arg[1024];

    snprintf(arg, 1023, browser->command, path);
    browser->pid = createProcess(arg, wnd);
    if (browser->pid != 0) {
        return TRUE;
    } else {
        return FALSE;
    }
}

static int
browserRemote(SBROWSER browser, const char *path)
{
    char	url[1024];
    char        arg[1024];
    pid_t	pid;

    snprintf(url, 1023, "file://%s", path);
    snprintf(arg, 1023, browser->remoteCommand, url);
    pid = createSimpleProcess(arg);
    return waitForProcess(pid);
}

static pid_t
createProcess(const char *cmdline, SWND wnd)
{
    pid_t       pid;
    int		fd[2];
    const char *argv[4];

    if (!cmdline || !wnd) return 0;

    argv[0] = "sh";
    argv[1] = "-c";
    argv[2] = cmdline;
    argv[3] = NULL;

    if (pipe(fd) != 0) return 0;

    pid = fork();
    switch(pid) {
      case -1:
	/* error */
	close(fd[0]);
	close(fd[1]);
	return 0;
      case 0:
	/* child process */
	if (fd[1] != STDERR_FILENO) {
	    dup2(fd[1], STDERR_FILENO); /* pipe output is now stderr */
	    close(fd[1]);	        /* close old pipe fds */
	}
	close(fd[0]);
	execv("/bin/sh", (char * const *) argv);
	exit(1);                /* if we get this far, it's an error */
      default:
	/* parent process */
	/* add fd[0] to list of inputs, so we can watch the shell's output */
	XtAppAddInput(TheAppContext, fd[0], (XtPointer) XtInputReadMask,
		      childInput, (XtPointer) wnd);
/*      close(fd[1]); */
	break;
    }
    return pid;
}

static pid_t
createSimpleProcess(const char *cmdline)
{
    pid_t       pid;
    const char *argv[4];

    if (!cmdline) return 0;

    argv[0] = "sh";
    argv[1] = "-c";
    argv[2] = cmdline;
    argv[3] = NULL;
    pid = fork();
    switch(pid) {
      case -1:
	/* error */
	return 0;
      case 0:
	/* child process */
	execv("/bin/sh", (char * const *) argv);
	exit(1);		/* if we get this far, it's an error */
      default:
	/* parent process */
	break;
    }
    return pid;
    return waitForProcess(pid);
}

static void
childInput(XtPointer closure , int *source, XtInputId *id)
{
    char                buf[1024];
    int                 bytes;

    bytes = read(*source, buf, 1023);
    buf[bytes] = '\0';
    swMessageBox((SWND) closure, buf, "Preview", SW_MB_OK, SW_MB_WARNING);
}

static int
waitForProcess(pid_t pid)
{
    int		status;

    for(;;) {
	if (waitpid(pid, &status, 0) == -1) {
	    if (errno != EINTR)
		return 0;
	} else {
	    return WIFEXITED(status) && WEXITSTATUS(status) == 0;
	}
    }
}
